@@ -37,6 +37,7 @@
 #include "hd-gtk-utils.h"
 #include "hd-gtk-style.h"
 #include "hd-transition.h"
+#include "hd-util.h"
 #include "hd-task-navigator.h"
 
 #include <matchbox/theme-engines/mb-wm-theme-png.h>
@@ -172,16 +173,24 @@ struct _HdTitleBarPrivate
   /* Stretched image for the title background */
   ClutterActor          *title_bg;
   ClutterLabel          *title;
+  /* The title to be used when in HDRM_STATE_LOADING */
+  gchar                 *loading_title;
   /* Pulsing animation for switcher */
   ClutterTimeline       *switcher_timeline;
   /* progress indicator */
   ClutterTimeline       *progress_timeline;
   ClutterActor          *progress_texture;
 
+  /* The current state set by set_state = next state we will go to once the
+   * idle callback is called */
   HdTitleBarVisEnum      state;
+  /* The actual state we are in right at this second */
+  HdTitleBarVisEnum      current_state;
   /* Do we show the menu indicator? This isn't included in state because
    * it could be unwillingly cleared by calls to set_state */
   gboolean               has_menu_indicator;
+  /* GSource id of the delayed hd_title_bar_update_idle() callback. */
+  guint                  update_title_bar;
 };
 
 /* HdHomeThemeButtonBack, MBWMDecorButtonClose */
@@ -206,7 +215,10 @@ on_switcher_timeline_new_frame(ClutterTimeline *timeline,
 static void
 hd_title_bar_set_full_width(HdTitleBar *bar, gboolean full_size);
 static void hd_title_bar_set_button_positions(HdTitleBar *bar);
-
+static void hd_title_bar_set_title (HdTitleBar *bar,
+                                    const char *title,
+                                    gboolean has_markup,
+                                    gboolean waiting);
 /* ------------------------------------------------------------------------- */
 
 /* One pulse is breathe in or breathe out.  The animation takes two
@@ -226,6 +238,9 @@ enum
   CLICKED_TOP_LEFT,
   PRESS_TOP_LEFT,
   LEAVE_TOP_LEFT,
+  CLICKED_TOP_RIGHT,
+  PRESS_TOP_RIGHT,
+  LEAVE_TOP_RIGHT,
   LAST_SIGNAL
 };
 static guint signals[LAST_SIGNAL] = { 0, };
@@ -337,14 +352,8 @@ hd_title_bar_init (HdTitleBar *bar)
   /* TODO: setup BTN_SWITCHER_HIGHLIGHT for adding here... */
 
   hd_title_bar_add_left_signals(bar, priv->buttons[BTN_SWITCHER]);
-  hd_render_manager_set_button (HDRM_BUTTON_TASK_NAV,
-                                priv->buttons[BTN_SWITCHER]);
   hd_title_bar_add_left_signals(bar, priv->buttons[BTN_LAUNCHER]);
-  hd_render_manager_set_button (HDRM_BUTTON_LAUNCHER,
-                                priv->buttons[BTN_LAUNCHER]);
   hd_title_bar_add_right_signals(bar, priv->buttons[BTN_BACK]);
-  hd_render_manager_set_button (HDRM_BUTTON_BACK,
-                                priv->buttons[BTN_BACK]);
 
   hd_title_bar_add_left_signals(bar, priv->buttons[BTN_MENU]);
   hd_title_bar_add_right_signals(bar, priv->buttons[BTN_DONE]);
@@ -380,6 +389,8 @@ hd_title_bar_init (HdTitleBar *bar)
                                                     &progress_geo);
     clutter_container_add_actor(CLUTTER_CONTAINER(bar),
                                 priv->progress_texture);
+    clutter_actor_set_visibility_detect(CLUTTER_ACTOR(priv->progress_texture),
+                                        TRUE);
     clutter_actor_set_size(priv->progress_texture,
                 HD_THEME_IMG_PROGRESS_SIZE, HD_THEME_IMG_PROGRESS_SIZE);
     clutter_actor_hide(priv->progress_texture);
@@ -425,8 +436,8 @@ status_area_is_visible (void)
 gint
 hd_title_bar_get_button_width(HdTitleBar *bar) {
   HdTitleBarPrivate *priv = bar->priv;
-  if (priv->state & HDTB_VIS_SMALL_BUTTONS)
-    return 80;
+  if (priv->current_state & HDTB_VIS_SMALL_BUTTONS)
+    return HD_COMP_MGR_TOP_RIGHT_BTN_WIDTH_SMALL;
   else
     /* TODO someday, in some cases this should return the TOP_RIGHT_BIN_WIDTH */
     return HD_COMP_MGR_TOP_LEFT_BTN_WIDTH;
@@ -438,6 +449,11 @@ hd_title_bar_dispose (GObject *obj)
   HdTitleBarPrivate *priv = HD_TITLE_BAR(obj)->priv;
   gint i;
 
+  if (priv->loading_title)
+    {
+      g_free(priv->loading_title);
+      priv->loading_title = 0;
+    }
   if (priv->progress_timeline)
     clutter_timeline_stop(priv->progress_timeline);
   for (i=0;i<BTN_COUNT;i++)
@@ -492,10 +508,31 @@ hd_title_bar_class_init (HdTitleBarClass *klass)
                       0, NULL, NULL,
                       g_cclosure_marshal_VOID__VOID,
                       G_TYPE_NONE, 0);
+  signals[CLICKED_TOP_RIGHT] =
+      g_signal_new ("clicked-top-right",
+                    G_TYPE_FROM_CLASS (klass),
+                    G_SIGNAL_RUN_LAST,
+                    0, NULL, NULL,
+                    g_cclosure_marshal_VOID__VOID,
+                    G_TYPE_NONE, 0);
+  signals[PRESS_TOP_RIGHT] =
+        g_signal_new ("press-top-right",
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      0, NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
+  signals[LEAVE_TOP_RIGHT] =
+        g_signal_new ("leave-top-right",
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      0, NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
 }
 
-void hd_title_bar_set_state(HdTitleBar *bar,
-                            HdTitleBarVisEnum button)
+static void hd_title_bar_set_state_real(HdTitleBar *bar,
+                                        HdTitleBarVisEnum button)
 {
   HdTitleBarPrivate *priv;
   if (!HD_IS_TITLE_BAR(bar))
@@ -602,6 +639,51 @@ void hd_title_bar_set_state(HdTitleBar *bar,
     }
 }
 
+#ifndef G_DEBUG_DISABLE
+# define PRINT_STATE(state, this)       \
+  if (((state) & this) == this)         \
+    g_debug("  " G_STRINGIFY(this))
+void hd_title_bar_print_state(HdTitleBar *bar, gboolean current)
+{
+  HdTitleBarVisEnum state;
+
+  if (current)
+    {
+      state = bar->priv->current_state;
+      g_debug("HdTitleBar::current_state:");
+    }
+  else
+    {
+      state = bar->priv->state;
+      g_debug("HdTitleBar::state:");
+    }
+
+  PRINT_STATE(state, HDTB_VIS_BTN_LAUNCHER);
+  PRINT_STATE(state, HDTB_VIS_BTN_SWITCHER);
+  PRINT_STATE(state, HDTB_VIS_BTN_MENU);
+  PRINT_STATE(state, HDTB_VIS_BTN_BACK);
+  PRINT_STATE(state, HDTB_VIS_BTN_CLOSE);
+  PRINT_STATE(state, HDTB_VIS_BTN_DONE);
+  PRINT_STATE(state, HDTB_VIS_FULL_WIDTH);
+  PRINT_STATE(state, HDTB_VIS_BTN_SWITCHER_HIGHLIGHT);
+  PRINT_STATE(state, HDTB_VIS_FOREGROUND);
+  PRINT_STATE(state, HDTB_VIS_SMALL_BUTTONS);
+  PRINT_STATE(state, HDTB_VIS_BTN_LEFT_MASK);
+  PRINT_STATE(state, HDTB_VIS_BTN_RIGHT_MASK);
+}
+# undef PRINT_STATE
+#endif
+
+void hd_title_bar_set_state(HdTitleBar *bar,
+                            HdTitleBarVisEnum button)
+{
+  if (!HD_IS_TITLE_BAR(bar))
+    return;
+
+  bar->priv->state = button;
+  hd_title_bar_update(bar);
+}
+
 HdTitleBarVisEnum hd_title_bar_get_state(HdTitleBar *bar)
 {
   HdTitleBarPrivate *priv;
@@ -807,13 +889,16 @@ hd_title_bar_set_for_edit_mode(HdTitleBar *bar)
 
   if (!HD_IS_TITLE_BAR(bar))
     return;
+
   priv = bar->priv;
 
+  priv->has_menu_indicator = TRUE;
+
   state |= HDTB_VIS_BTN_MENU;
   state |= HDTB_VIS_BTN_DONE;
   state |= HDTB_VIS_FULL_WIDTH;
 
-  hd_title_bar_set_state(bar, state);
+  hd_title_bar_set_state_real(bar, state);
 
   title = dgettext ("maemo-af-desktop", "home_ti_desktop_menu");
   hd_title_bar_set_title (bar, title, FALSE, FALSE);
@@ -832,7 +917,8 @@ hd_title_bar_set_for_loading(HdTitleBar *bar)
   state |= HDTB_VIS_FULL_WIDTH;
   state &= ~HDTB_VIS_FOREGROUND;
 
-  hd_title_bar_set_state(bar, state);
+  hd_title_bar_set_state_real(bar, state);
+  hd_title_bar_set_title (bar, priv->loading_title, FALSE, TRUE);
 }
 
 static void
@@ -884,7 +970,7 @@ hd_title_bar_set_window(HdTitleBar *bar, MBWindowManagerClient *client)
       HdTitleBarVisEnum state;
       /* No north decor found, or no client */
       /* we have nothing, make sure we're back to normal */
-      clutter_actor_hide(CLUTTER_ACTOR(priv->title));
+      hd_title_bar_set_title (bar, NULL, FALSE, FALSE);
       if (hd_render_manager_get_state() &
 	  (HDRM_STATE_HOME_EDIT | HDRM_STATE_HOME_EDIT_DLG))
         /* in Home edit mode we can have back button */
@@ -892,7 +978,7 @@ hd_title_bar_set_window(HdTitleBar *bar, MBWindowManagerClient *client)
       else
 	state = hd_title_bar_get_state(bar) & ~(HDTB_VIS_FULL_WIDTH |
 			         HDTB_VIS_BTN_CLOSE | HDTB_VIS_BTN_BACK);
-      hd_title_bar_set_state(bar, state);
+      hd_title_bar_set_state_real(bar, state);
       return;
     }
 
@@ -925,7 +1011,7 @@ hd_title_bar_set_window(HdTitleBar *bar, MBWindowManagerClient *client)
   /* Also set us to be full width */
   state |= HDTB_VIS_FULL_WIDTH;
 
-  hd_title_bar_set_state(bar, state);
+  hd_title_bar_set_state_real(bar, state);
   hd_title_bar_right_pressed(bar, pressed);
 }
 
@@ -958,10 +1044,10 @@ static gint hd_title_bar_get_end_of_title(HdTitleBar *bar,
   return x;
 }
 
-void hd_title_bar_set_title (HdTitleBar *bar,
-                             const char *title,
-                             gboolean has_markup,
-                             gboolean waiting)
+static void hd_title_bar_set_title (HdTitleBar *bar,
+                                    const char *title,
+                                    gboolean has_markup,
+                                    gboolean waiting)
 {
   HdTitleBarPrivate *priv;
   if (!HD_IS_TITLE_BAR(bar))
@@ -1025,8 +1111,9 @@ void hd_title_bar_set_title (HdTitleBar *bar,
     }
 
   /* Only show the indicator if we are asked to, and we're not
-   * showing the waiting indicator. */
-  if (priv->has_menu_indicator && !waiting)
+   * showing the waiting indicator. Also don't show it if we have
+   * no title, as we're probably just waiting in a loading screen. */
+  if (priv->has_menu_indicator && title && !waiting)
      {
        gint x = hd_title_bar_get_end_of_title(bar,
                                               HD_THEME_IMG_MENU_INDICATOR_SIZE,
@@ -1047,28 +1134,29 @@ void hd_title_bar_set_title (HdTitleBar *bar,
      clutter_actor_hide(priv->buttons[BTN_MENU_INDICATOR]);
 }
 
-void
-hd_title_bar_update(HdTitleBar *bar, MBWMCompMgr *wmcm)
+void hd_title_bar_set_loading_title   (HdTitleBar *bar,
+                                       const char *title)
 {
-  MBWindowManagerClient *client = 0;
-  if (STATE_IS_APP(hd_render_manager_get_state()))
-    {
-      /* find the topmost application above the desktop, and use this
-       * for setting our title - or have NULL */
-      client = wmcm->wm->stack_top;
-      while (client)
-        {
-          MBWMClientType c_type = MB_WM_CLIENT_CLIENT_TYPE(client);
-          if (c_type == MBWMClientTypeApp)
-            break;
-          if (c_type == MBWMClientTypeDesktop)
-            {
-              client = 0;
-              break;
-            }
-          client = client->stacked_below;
-        }
-    }
+  HdTitleBarPrivate *priv;
+  if (!HD_IS_TITLE_BAR(bar))
+    return;
+  priv = bar->priv;
+
+  if (priv->loading_title)
+    g_free(priv->loading_title);
+  priv->loading_title = g_strdup(title);
+
+  hd_title_bar_update(bar);
+}
+
+static gboolean
+hd_title_bar_update_idle(HdTitleBar *bar)
+{
+  HdTitleBarPrivate *priv = bar->priv;
+  HdTitleBarVisEnum old_state;
+
+  old_state = priv->current_state;
+  priv->current_state = priv->state;
 
   if (STATE_IS_EDIT_MODE(hd_render_manager_get_state()))
     {
@@ -1082,9 +1170,76 @@ hd_title_bar_update(HdTitleBar *bar, MBWMCompMgr *wmcm)
     }
   else
     {
-      hd_title_bar_set_window(bar, client);
-      clutter_actor_set_reactive (bar->priv->title_bg, FALSE);
+      MBWindowManagerClient *c;
+
+      if (STATE_IS_APP (hd_render_manager_get_state ()))
+        {
+          extern MBWindowManager *hd_mb_wm;
+          HdCompMgr *cmgr = HD_COMP_MGR(hd_mb_wm->comp_mgr);
+          HdCompMgrClient *cclient = hd_comp_mgr_get_current_client(cmgr);
+          c = cclient ? MB_WM_COMP_MGR_CLIENT(cclient)->wm_client : NULL;
+        }
+      else
+        c = NULL;
+
+      hd_title_bar_set_window(bar, c);
+      clutter_actor_set_reactive(bar->priv->title_bg, FALSE);
     }
+
+  /* If anything that might change the position of the status area to change
+   * has been modified, re-place titlebar elements now */
+  if ((priv->current_state & (HDTB_VIS_SMALL_BUTTONS|HDTB_VIS_BTN_LEFT_MASK)) !=
+      (old_state           & (HDTB_VIS_SMALL_BUTTONS|HDTB_VIS_BTN_LEFT_MASK)))
+    hd_render_manager_place_titlebar_elements();
+
+  /* Unfortunately, as we have updated title visibility of buttons and so forth,
+   * we must now update the input viewport */
+  hd_render_manager_set_input_viewport();
+
+  /* This is only for this idle callback, so don't leave it dangling */
+  priv->update_title_bar = 0;
+  return FALSE;
+}
+
+void
+hd_title_bar_update(HdTitleBar *bar)
+{
+  HdTitleBarPrivate *priv = bar->priv;
+
+  if (!priv->update_title_bar)
+    /* Add idle callback. This MUST be higher priority than Clutter timelines
+     * (D+30) or we won't set our title bar up correctly until after any running
+     * transitions have stopped. */
+    priv->update_title_bar = g_idle_add_full(G_PRIORITY_DEFAULT+20,
+                                  (GSourceFunc)hd_title_bar_update_idle,
+                                  bar, NULL);
+}
+
+void
+hd_title_bar_update_now(HdTitleBar *bar)
+{
+  HdTitleBarPrivate *priv = bar->priv;
+  HdTitleBarVisEnum prev_button_size;
+
+  if (priv->update_title_bar)
+    {
+      g_source_remove(priv->update_title_bar);
+      priv->update_title_bar = 0;
+    }
+
+  /* We're going to portrait but we're not quite there yet,
+   * so the title bar must look landscape but the title must
+   * be the new portrait application's. */
+  prev_button_size = priv->state & HDTB_VIS_SMALL_BUTTONS;
+  if (hd_transition_is_rotating())
+    {
+      if (STATE_IS_PORTRAIT(hd_render_manager_get_previous_state()))
+        priv->state |= HDTB_VIS_SMALL_BUTTONS;
+      else
+        priv->state &= ~HDTB_VIS_SMALL_BUTTONS;
+    }
+  hd_title_bar_update_idle(bar);
+  priv->state = (priv->state & ~HDTB_VIS_SMALL_BUTTONS) | prev_button_size;
 }
 
 /* Is the given decor one we should consider for a title bar? */
@@ -1168,6 +1323,8 @@ static void hd_title_bar_set_button_positions(HdTitleBar *bar)
 
 /* ------------------------------------------------------------------------- */
 
+extern gboolean hd_dbus_display_is_off;
+
 static void
 on_switcher_timeline_new_frame(ClutterTimeline *timeline,
                                gint frame_num, HdTitleBar *bar)
@@ -1175,24 +1332,22 @@ on_switcher_timeline_new_frame(ClutterTimeline *timeline,
   HdTitleBarPrivate *priv;
   float amt;
   gint opacity;
-  ClutterActor *actor, *stage;
-  gboolean is_visible = TRUE;
 
   if (!HD_IS_TITLE_BAR(bar))
     return;
   priv = bar->priv;
 
-  /* Only get this to fire a redraw if it is visible... fixes bug 113278.
-   * Clutter should do this, but it only checks the actor for visibility,
-   * not its parents. Changing clutter at this stage might risk regressing
-   * things too much... */
-  stage = clutter_actor_get_stage(CLUTTER_ACTOR(bar));
-  for (actor=CLUTTER_ACTOR(bar);actor;actor=clutter_actor_get_parent(actor))
-    if (actor!=stage && !CLUTTER_ACTOR_IS_VISIBLE(actor))
-      is_visible = FALSE;
+  if (hd_dbus_display_is_off)
+    {
+      /* skip the animation */
+      clutter_actor_set_opacity(priv->buttons[BTN_SWITCHER_HIGHLIGHT], 255);
+      return;
+    }
 
-  if (!is_visible)
-    clutter_actor_set_allow_redraw(CLUTTER_ACTOR(bar), FALSE);
+  /* Only get this to fire a redraw if it is visible... fixes bug 113278.
+     (and now only update the actual area using
+      hd_util_partial_redraw_if_possible...) */
+  clutter_actor_set_allow_redraw(CLUTTER_ACTOR(bar), FALSE);
 
   amt =  (float)clutter_timeline_get_progress(timeline)
               * HD_TITLE_BAR_SWITCHER_PULSE_NPULSES / 2;
@@ -1202,8 +1357,8 @@ on_switcher_timeline_new_frame(ClutterTimeline *timeline,
       clutter_actor_set_opacity(priv->buttons[BTN_SWITCHER_HIGHLIGHT], opacity);
     }
 
-  if (!is_visible)
-      clutter_actor_set_allow_redraw(CLUTTER_ACTOR(bar), TRUE);
+  hd_util_partial_redraw_if_possible(priv->buttons[BTN_SWITCHER_HIGHLIGHT], 0);
+  clutter_actor_set_allow_redraw(CLUTTER_ACTOR(bar), TRUE);
 }
 
 /* Realign all right-aligned buttons when the screen size changes. */
@@ -1237,6 +1392,16 @@ hd_title_bar_top_left_leave (HdTitleBar *bar)
   g_signal_emit (bar, signals[LEAVE_TOP_LEFT], 0);
 }
 
+static void
+hd_title_bar_top_right_clicked (HdTitleBar *bar)
+{
+  /* For if we ever add the back button back into launcher */
+  if (hd_render_manager_get_state() == HDRM_STATE_LAUNCHER)
+    hd_launcher_back_button_clicked();
+
+  g_signal_emit (bar, signals[CLICKED_TOP_RIGHT], 0);
+}
+
 static void
 hd_title_bar_top_right_press (HdTitleBar *bar)
 {
@@ -1245,7 +1410,7 @@ hd_title_bar_top_right_press (HdTitleBar *bar)
   if (hd_render_manager_get_state() == HDRM_STATE_HOME_EDIT)
     hd_render_manager_set_state(HDRM_STATE_HOME);
 
-  g_signal_emit (bar, signals[PRESS_TOP_LEFT], 0);
+  g_signal_emit (bar, signals[PRESS_TOP_RIGHT], 0);
 }
 
 static void
@@ -1253,7 +1418,7 @@ hd_title_bar_top_right_leave (HdTitleBar *bar)
 {
   hd_title_bar_right_pressed(bar, FALSE);
 
-  g_signal_emit (bar, signals[LEAVE_TOP_LEFT], 0);
+  g_signal_emit (bar, signals[LEAVE_TOP_RIGHT], 0);
 }
 
 static void
@@ -1275,6 +1440,9 @@ static void
 hd_title_bar_add_right_signals(HdTitleBar *bar, ClutterActor *actor)
 {
   clutter_actor_set_reactive(actor, TRUE);
+  g_signal_connect_swapped (actor, "button-release-event",
+                            G_CALLBACK (hd_title_bar_top_right_clicked),
+                            bar);
   g_signal_connect_swapped (actor, "button-press-event",
                             G_CALLBACK (hd_title_bar_top_right_press),
                             bar);
