@@ -48,6 +48,7 @@
 #include "hd-launcher-tree.h"
 #include "home/hd-render-manager.h"
 #include "hd-transition.h"
+#include "hd-wm.h"
 
 #undef  G_LOG_DOMAIN
 #define G_LOG_DOMAIN "hd-app-mgr"
@@ -121,7 +122,6 @@ struct _HdAppMgrPrivate
   GConfClient *gconf_client;
   gboolean portrait;
   gboolean unlocked;
-  gboolean display_on;
   gboolean slide_closed;
   gboolean disable_callui;
   gboolean accel_enabled;
@@ -220,7 +220,6 @@ gboolean hd_app_mgr_hibernate (HdRunningApp *app);
 
 static gboolean hd_app_mgr_service_top (const gchar *service,
                                         const gchar *param);
-static gboolean  hd_app_mgr_execute (const gchar *exec, GPid *pid);
 
 void hd_app_mgr_prestartable     (HdRunningApp *app, gboolean prestartable);
 
@@ -242,7 +241,7 @@ hd_app_mgr_setup_launch (size_t high_pages,
                          size_t nr_decay_pages,
                          size_t *launch_required_pages);
 static gboolean hd_app_mgr_can_launch   (HdLauncherApp *launcher);
-static gboolean hd_app_mgr_can_prestart (void);
+static gboolean hd_app_mgr_can_prestart (HdLauncherApp *launcher);
 static void     hd_app_mgr_hdrm_state_change (gpointer hdrm,
                                               GParamSpec *pspec,
                                               HdAppMgrPrivate *priv);
@@ -264,9 +263,7 @@ static void hd_app_mgr_gconf_value_changed (GConfClient *client,
                                             guint cnxn_id,
                                             GConfEntry *entry,
                                             gpointer user_data);
-static void hd_app_mgr_mce_activate_accel  (void);
 static gboolean hd_app_mgr_show_callui_cb (gpointer data);
-static void hd_app_mgr_check_show_callui (void);
 
 static void     hd_app_mgr_request_app_pid (HdRunningApp *app);
 static gboolean hd_app_mgr_loading_timeout (HdRunningApp *app);
@@ -396,7 +393,8 @@ hd_app_mgr_init (HdAppMgr *self)
   for (int i = 0; i < NUM_QUEUES; i++)
     priv->queues[i] = g_queue_new ();
 
-  priv->tree = hd_launcher_tree_new (NULL);
+  priv->tree = hd_launcher_tree_new ();
+  hd_launcher_tree_ensure_user_menu ();
   g_signal_connect (priv->tree, "finished",
                     G_CALLBACK (hd_app_mgr_populate_tree_finished),
                     self);
@@ -404,7 +402,6 @@ hd_app_mgr_init (HdAppMgr *self)
 
   /* NOTE: Can we assume this when we start up? */
   priv->unlocked = TRUE;
-  priv->display_on = TRUE;
   priv->gconf_client = gconf_client_get_default ();
   if (priv->gconf_client)
     {
@@ -421,7 +418,7 @@ hd_app_mgr_init (HdAppMgr *self)
                                                     GCONF_DISABLE_CALLUI_KEY,
                                                     NULL);
       /* We don't call
-      hd_app_mgr_mce_activate_accel ();
+      hd_app_mgr_mce_activate_accel_if_needed ();
       here because hdrm is not ready yet. */
       gconf_client_add_dir (priv->gconf_client, GCONF_DISABLE_CALLUI_DIR,
                             GCONF_CLIENT_PRELOAD_NONE, NULL);
@@ -927,7 +924,7 @@ hd_app_mgr_start (HdRunningApp *app)
       if (exec)
         {
           GPid pid = 0;
-          result = hd_app_mgr_execute (exec, &pid);
+          result = hd_app_mgr_execute (exec, &pid, FALSE);
           if (result)
             {
               hd_running_app_set_pid (app, pid);
@@ -1347,8 +1344,8 @@ _hd_app_mgr_child_setup(gpointer user_data)
   }
 }
 
-static gboolean
-hd_app_mgr_execute (const gchar *exec, GPid *pid)
+gboolean
+hd_app_mgr_execute (const gchar *exec, GPid *pid, gboolean auto_reap)
 {
   gboolean res = FALSE;
   gchar *space = strchr (exec, ' ');
@@ -1381,7 +1378,7 @@ hd_app_mgr_execute (const gchar *exec, GPid *pid)
 
   res = g_spawn_async (NULL,
                        argv, NULL,
-                       G_SPAWN_DO_NOT_REAP_CHILD,
+                       auto_reap ? 0 : G_SPAWN_DO_NOT_REAP_CHILD,
                        _hd_app_mgr_child_setup, NULL,
                        pid,
                        NULL);
@@ -1539,7 +1536,7 @@ hd_app_mgr_can_launch (HdLauncherApp *launcher)
   return !priv->lowmem;
 }
 
-static gboolean hd_app_mgr_can_prestart (void)
+static gboolean hd_app_mgr_can_prestart (HdLauncherApp *launcher)
 {
   HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (hd_app_mgr_get ());
   if (priv->prestart_mode == PRESTART_ALWAYS)
@@ -1547,6 +1544,9 @@ static gboolean hd_app_mgr_can_prestart (void)
   else if (priv->prestart_mode == PRESTART_NEVER)
     return FALSE;
 
+  if (hd_launcher_app_get_ignore_load (launcher))
+    return TRUE;
+
   if (!hd_app_mgr_check_loadavg ())
     return FALSE;
 
@@ -1570,7 +1570,7 @@ hd_app_mgr_hdrm_state_change (gpointer hdrm,
     }
 
   /* Also check if we should enable the accelerometer. */
-  hd_app_mgr_mce_activate_accel ();
+  hd_app_mgr_mce_activate_accel_if_needed (TRUE);
 }
 
 static void
@@ -1652,10 +1652,12 @@ hd_app_mgr_state_check_loop (gpointer data)
       )
     {
       /* We make this tests here to loop even if we can't prestart right now.*/
-      if (!priv->prestarting && hd_app_mgr_can_prestart ())
+      if (!priv->prestarting)
         {
           HdRunningApp *app = g_queue_peek_head (priv->queues[QUEUE_PRESTARTABLE]);
-          hd_app_mgr_prestart (app);
+          HdLauncherApp *launcher = hd_running_app_get_launcher_app (app);
+          if (launcher && hd_app_mgr_can_prestart (launcher))
+            hd_app_mgr_prestart (app);
         }
       if (!g_queue_is_empty (priv->queues[QUEUE_PRESTARTABLE]))
         loop = TRUE;
@@ -1787,7 +1789,7 @@ hd_app_mgr_dbus_launch_app (HdAppMgr *self, const gchar *id)
 {
   HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (self);
   HdLauncherItem *item;
-  HdLauncherApp *app = NULL;
+  HdLauncherApp *app;
 
   if (!id || !strlen(id))
     { /* If no ID was specified, all we want to do is trigger the
@@ -1801,15 +1803,11 @@ hd_app_mgr_dbus_launch_app (HdAppMgr *self, const gchar *id)
     }
 
   item = hd_launcher_tree_find_item (priv->tree, id);
-
-  if (!item)
-    return FALSE;
-  if (hd_launcher_item_get_item_type (item) != HD_APPLICATION_LAUNCHER)
-    return FALSE;
-
-  app = HD_LAUNCHER_APP (item);
-
-  return hd_app_mgr_launch (app);
+  if (!item || hd_launcher_item_get_item_type (item) != HD_APPLICATION_LAUNCHER)
+    app = hd_launcher_tree_find_app_by_service (priv->tree, id);
+  else
+    app = HD_LAUNCHER_APP (item);
+  return app ? hd_app_mgr_launch (app) : FALSE;
 }
 
 gboolean
@@ -1835,13 +1833,17 @@ static gboolean
 _hd_app_mgr_should_show_callui ()
 {
   HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (hd_app_mgr_get ());
+  extern MBWindowManager *hd_mb_wm;
+  extern gboolean hd_dbus_display_is_off;
 
   if (STATE_SHOW_CALLUI (hd_render_manager_get_state ()) &&
+      priv->accel_enabled &&
       priv->portrait &&
       priv->unlocked &&
-      priv->display_on &&
+      !hd_dbus_display_is_off &&
       priv->slide_closed &&
-      !priv->disable_callui)
+      !priv->disable_callui &&
+      !hd_wm_has_modal_blockers (hd_mb_wm))
     {
       return TRUE;
     }
@@ -1866,7 +1868,7 @@ hd_app_mgr_show_callui_cb (gpointer data)
  * - Display on.
  * - Slide closed.
  */
-static void
+void
 hd_app_mgr_check_show_callui (void)
 {
   if (_hd_app_mgr_should_show_callui ())
@@ -1918,13 +1920,30 @@ _hd_app_mgr_dbus_check_value (DBusMessage *msg,
   return FALSE;
 }
 
+static void
+hd_app_mgr_update_portraitness(HdAppMgr *self)
+{
+  HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (self);
+  HdCompMgr *hmgr = hd_comp_mgr_get ();
+  // what about priv->unlocked/priv->display_on ?
+
+  if (!hmgr)
+    /* We're in some initialization state. */
+    return;
+  hd_comp_mgr_set_pip_flags (hmgr,
+      priv->accel_enabled,
+      priv->portrait && priv->slide_closed);
+  hd_comp_mgr_portrait_or_not_portrait (MB_WM_COMP_MGR (hmgr), NULL);
+}
+
 static DBusHandlerResult
 hd_app_mgr_dbus_signal_handler (DBusConnection *conn,
                            DBusMessage *msg,
                            void *data)
 {
   gboolean changed = TRUE;
-  HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (HD_APP_MGR (data));
+  HdAppMgr *self = HD_APP_MGR (data);
+  HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (self);
 
   if (dbus_message_is_signal (msg,
                               LOWMEM_ON_SIGNAL_INTERFACE,
@@ -1957,13 +1976,6 @@ hd_app_mgr_dbus_signal_handler (DBusConnection *conn,
           priv->unlocked = _hd_app_mgr_dbus_check_value (msg,
                                            MCE_DEVICE_UNLOCKED);
         }
-      else if (dbus_message_is_signal (msg,
-                                       MCE_SIGNAL_IF,
-                                       MCE_DISPLAY_SIG))
-        {
-          priv->display_on = _hd_app_mgr_dbus_check_value (msg,
-                                           MCE_DISPLAY_ON_STRING);
-        }
       else if (dbus_message_is_signal (msg,
                                   MCE_SIGNAL_IF,
                                   MCE_DEVICE_ORIENTATION_SIG))
@@ -1972,6 +1984,8 @@ hd_app_mgr_dbus_signal_handler (DBusConnection *conn,
                                            MCE_ORIENTATION_PORTRAIT);
           if (priv->portrait)
             hd_app_mgr_check_show_callui ();
+
+          hd_app_mgr_update_portraitness(self);
         }
     }
 
@@ -1984,16 +1998,22 @@ hd_app_mgr_dbus_signal_handler (DBusConnection *conn,
 /* Activate the accelerometer when
  * - The user has activated rotate-to-callui.
  * - HDRM is in a state that shows callui.
+ * - We are showing an app, and all visible windows support portrait mode
  */
-static void
-hd_app_mgr_mce_activate_accel  ()
+void
+hd_app_mgr_mce_activate_accel_if_needed (gboolean update_portraitness)
 {
+  extern gboolean hd_dbus_tklock_on;
   HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (the_app_mgr);
   DBusConnection *conn = NULL;
   DBusMessage *msg = NULL;
+  gboolean activate = !hd_dbus_tklock_on;
 
-  gboolean activate = !priv->disable_callui &&
-                      STATE_SHOW_CALLUI (hd_render_manager_get_state ());
+  if (activate)
+    activate = (!priv->disable_callui
+                && STATE_SHOW_CALLUI (hd_render_manager_get_state ()))
+      || (STATE_IS_APP(hd_render_manager_get_state ())
+          && hd_comp_mgr_can_be_portrait(hd_comp_mgr_get()));
   if (priv->accel_enabled == activate)
     return;
 
@@ -2034,21 +2054,39 @@ hd_app_mgr_mce_activate_accel  ()
     }
 
   dbus_message_set_auto_start (msg, TRUE);
-  dbus_message_set_no_reply (msg, TRUE);
-
-  if (!dbus_connection_send (conn, msg, NULL))
+  if (activate)
     {
-      g_warning ("%s: Couldn't send message.", __FUNCTION__);
+      DBusMessage *reply;
+
+      /* @reply will contain the current orientation */
+      if ((reply = dbus_connection_send_with_reply_and_block (
+                                          conn, msg, -1, NULL)) != NULL)
+        {
+          priv->portrait = _hd_app_mgr_dbus_check_value (reply,
+                                              MCE_ORIENTATION_PORTRAIT);
+          dbus_message_unref (reply);
+        }
+      else
+        g_warning ("%s: Couldn't send message.", __FUNCTION__);
+      dbus_message_unref (msg);
     }
   else
-    dbus_connection_flush(conn);
+    { /* Deactivate, expect no reply.  We can only deactivate in lscape. */
+      dbus_message_set_no_reply (msg, TRUE);
+      if (!dbus_connection_send (conn, msg, NULL))
+        g_warning ("%s: Couldn't send message.", __FUNCTION__);
+      dbus_message_unref (msg);
+      priv->portrait = FALSE;
+    }
 
-  dbus_message_unref (msg);
-
-  g_debug ("%s: %s", __FUNCTION__,
-           activate? "enabled" : "disabled");
+  g_debug ("%s: %s", __FUNCTION__, activate ? "enabled" : "disabled");
   priv->accel_enabled = activate;
-  return;
+
+  if (update_portraitness)
+    hd_app_mgr_update_portraitness(the_app_mgr);
+  else
+    hd_comp_mgr_set_pip_flags (hd_comp_mgr_get (), priv->accel_enabled,
+                               priv->portrait && priv->slide_closed);
 }
 
 static void
@@ -2057,7 +2095,8 @@ hd_app_mgr_gconf_value_changed (GConfClient *client,
                                 GConfEntry *entry,
                                 gpointer user_data)
 {
-  HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (HD_APP_MGR (user_data));
+  HdAppMgr *self = HD_APP_MGR (user_data);
+  HdAppMgrPrivate *priv = HD_APP_MGR_GET_PRIVATE (self);
   GConfValue *gvalue;
   gboolean value = FALSE;
 
@@ -2070,7 +2109,11 @@ hd_app_mgr_gconf_value_changed (GConfClient *client,
 
   if (!g_strcmp0 (gconf_entry_get_key (entry),
                   GCONF_SLIDE_OPEN_KEY))
-    priv->slide_closed = !value;
+    {
+      priv->slide_closed = !value;
+
+      hd_app_mgr_update_portraitness(self);
+    }
 
   if (!g_strcmp0 (gconf_entry_get_key (entry),
                   GCONF_DISABLE_CALLUI_KEY))
@@ -2078,7 +2121,7 @@ hd_app_mgr_gconf_value_changed (GConfClient *client,
       priv->disable_callui = value;
 
       /* Check if h-d needs to track the orientation. */
-      hd_app_mgr_mce_activate_accel ();
+      hd_app_mgr_mce_activate_accel_if_needed (TRUE);
     }
 
   return;
