@@ -50,6 +50,8 @@
 /* We don't use a font from the theme here because apparently there are none the
  * size we want, and we don't want to add another logical font to gtkrc. */
 
+#define HD_LAUNCHER_TILE_LONG_PRESS_DUR (1000)
+
 struct _HdLauncherTilePrivate
 {
   gchar *icon_name;
@@ -66,6 +68,7 @@ struct _HdLauncherTilePrivate
   float glow_radius; // radius of glow - loaded from transitions.ini
 
   /* We need to know if there's been scrolling. */
+  guint    press_timeout;
   gboolean is_pressed;
 };
 
@@ -80,6 +83,7 @@ enum
 enum
 {
   CLICKED,
+  LONG_CLICKED,
 
   LAST_SIGNAL
 };
@@ -104,6 +108,7 @@ static gboolean hd_launcher_tile_button_release (ClutterActor       *actor);
 static void hd_launcher_on_glow_frame(ClutterTimeline *timeline,
                                       gint frame_num,
                                       ClutterActor *actor);
+
 static void hd_launcher_tile_allocate (ClutterActor          *self,
                                        const ClutterActorBox *box,
                                        gboolean       absolute_origin_changed);
@@ -147,6 +152,14 @@ hd_launcher_tile_class_init (HdLauncherTileClass *klass)
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
+  launcher_tile_signals[LONG_CLICKED] =
+    g_signal_new (I_("long-clicked"),
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
 }
 
 static void
@@ -172,11 +185,11 @@ hd_launcher_tile_init (HdLauncherTile *tile)
    * It's good that we can make this actor a rectangle and see exactly
    * where the user is allowed to click - but to make it not draw anything
    * but still be selectable */
-  if (TRUE)
+  if (!hd_transition_get_int("blur", "turbo", 0))
     priv->click_area = clutter_group_new();
   else
     {
-      ClutterColor red = {0xFF, 0x00, 0x00, 0x3F};
+      ClutterColor red = {0x7F, 0x20, 0x20, 0xA0};
       priv->click_area = clutter_rectangle_new_with_color(&red);
     }
 
@@ -272,6 +285,7 @@ hd_launcher_tile_set_icon_name (HdLauncherTile *tile,
   HdLauncherTilePrivate *priv = HD_LAUNCHER_TILE_GET_PRIVATE (tile);
   GtkIconTheme *icon_theme;
   GtkIconInfo *info;
+  GdkPixbuf *pixbuf;
 
   if (priv->icon_name)
     {
@@ -292,7 +306,7 @@ hd_launcher_tile_set_icon_name (HdLauncherTile *tile,
 
   icon_theme = gtk_icon_theme_get_default();
   info = gtk_icon_theme_lookup_icon(icon_theme, priv->icon_name,
-                                    HD_LAUNCHER_TILE_ICON_SIZE,
+                                    HD_LAUNCHER_TILE_ICON_REAL_SIZE,
                                     GTK_ICON_LOOKUP_NO_SVG);
   if (info == NULL)
     {
@@ -300,24 +314,70 @@ hd_launcher_tile_set_icon_name (HdLauncherTile *tile,
       g_free (priv->icon_name);
       priv->icon_name = g_strdup (HD_LAUNCHER_DEFAULT_ICON);
       info = gtk_icon_theme_lookup_icon(icon_theme, priv->icon_name,
-                                        HD_LAUNCHER_TILE_ICON_SIZE,
+                                        HD_LAUNCHER_TILE_ICON_REAL_SIZE,
                                         GTK_ICON_LOOKUP_NO_SVG);
     }
   if (info == NULL)
     {
       g_warning ("%s: couldn't find icon %s\n", __FUNCTION__, priv->icon_name);
+      g_free (priv->icon_name);
+      priv->icon_name = NULL;
       return;
     }
 
   const gchar *fname = gtk_icon_info_get_filename(info);
-  priv->icon = clutter_texture_new_from_file(fname, NULL);
+  if (fname == NULL)
+    {
+      g_warning ("%s: couldn't get icon %s\n", __FUNCTION__, priv->icon_name);
+      g_free (priv->icon_name);
+      priv->icon_name = NULL;
+      return;
+    }
+
+  /* We must expand these images so there is a 1 pixel transparent
+   * border around them, or the glow effect won't work properly. We use
+   * gdk_pixbuf_new_from_file_at_size as the pixbuf pointed to by fname
+   * isn't actually guaranteed to be the correct size.  */
+  pixbuf = gdk_pixbuf_new_from_file_at_size(fname,
+      HD_LAUNCHER_TILE_ICON_REAL_SIZE, HD_LAUNCHER_TILE_ICON_REAL_SIZE, 0);
+  if (pixbuf)
+    {
+      gint w = gdk_pixbuf_get_width(pixbuf);
+      gint h = gdk_pixbuf_get_height(pixbuf);
+      GdkPixbuf *pixbufb = gdk_pixbuf_new(
+          GDK_COLORSPACE_RGB, TRUE, 8, w+2, h+2);
+
+      gdk_pixbuf_fill(pixbufb, 0);
+      gdk_pixbuf_copy_area(pixbuf, 0, 0, w, h,
+                           pixbufb, 1, 1);
+
+      priv->icon = clutter_texture_new();
+      clutter_texture_set_from_rgb_data(
+          CLUTTER_TEXTURE(priv->icon),
+          gdk_pixbuf_get_pixels(pixbufb),
+          gdk_pixbuf_get_has_alpha(pixbufb),
+          gdk_pixbuf_get_width(pixbufb),
+          gdk_pixbuf_get_height(pixbufb),
+          gdk_pixbuf_get_rowstride(pixbufb),
+          gdk_pixbuf_get_n_channels(pixbufb), 0, 0);
+      g_object_unref(pixbufb);
+    }
+  if (pixbuf)
+    g_object_unref(pixbuf);
+
+  if (!priv->icon)
+  {
+    g_warning ("%s: couldn't create texture for %s\n", __FUNCTION__, fname);
+    g_free (priv->icon_name);
+    priv->icon_name = NULL;
+    return;
+  }
 
   clutter_actor_set_size (priv->icon,
       HD_LAUNCHER_TILE_ICON_SIZE,
       HD_LAUNCHER_TILE_ICON_SIZE);
   clutter_actor_set_position (priv->icon,
-      (HD_LAUNCHER_TILE_WIDTH - HD_LAUNCHER_TILE_ICON_SIZE) / 2,
-      HILDON_MARGIN_HALF);
+      (HD_LAUNCHER_TILE_WIDTH - HD_LAUNCHER_TILE_ICON_SIZE) / 2, 0);
   clutter_container_add_actor (CLUTTER_CONTAINER(tile), priv->icon);
 
   gtk_icon_info_free(info);
@@ -332,7 +392,6 @@ hd_launcher_tile_set_icon_name (HdLauncherTile *tile,
         HD_LAUNCHER_TILE_GLOW_SIZE);
   clutter_actor_set_position (CLUTTER_ACTOR(priv->icon_glow),
         (HD_LAUNCHER_TILE_WIDTH - HD_LAUNCHER_TILE_GLOW_SIZE) / 2,
-        HILDON_MARGIN_HALF +
         (HD_LAUNCHER_TILE_ICON_SIZE - HD_LAUNCHER_TILE_GLOW_SIZE) / 2);
   clutter_container_add_actor (CLUTTER_CONTAINER(tile),
                                CLUTTER_ACTOR(priv->icon_glow));
@@ -379,7 +438,7 @@ hd_launcher_tile_set_text (HdLauncherTile *tile,
   clutter_label_set_line_wrap_mode (CLUTTER_LABEL (priv->label),
                                     PANGO_WRAP_CHAR);
 
-  label_height = HD_LAUNCHER_TILE_HEIGHT - (64 + (HILDON_MARGIN_HALF * 2));
+  label_height = HD_LAUNCHER_TILE_HEIGHT - (64 + HILDON_MARGIN_HALF);
 
   clutter_actor_get_preferred_width (priv->label,
     CLUTTER_UNITS_FROM_DEVICE(label_height),
@@ -476,7 +535,8 @@ hd_launcher_tile_set_glow(HdLauncherTile *tile, gboolean glow)
   hd_gtk_style_get_text_color(HD_GTK_BUTTON_SINGLETON, GTK_STATE_NORMAL,
                               &glow_col);
   glow_col.alpha = (int)(glow_col.alpha * glow_brightness);
-  tidy_highlight_set_color(priv->icon_glow, &glow_col);
+  if (priv->icon_glow)
+    tidy_highlight_set_color(priv->icon_glow, &glow_col);
   /* load our glow radius */
   priv->glow_radius = hd_transition_get_double("launcher_glow", "radius", 8);
 
@@ -484,6 +544,21 @@ hd_launcher_tile_set_glow(HdLauncherTile *tile, gboolean glow)
   clutter_timeline_start(priv->glow_timeline);
 }
 
+static gboolean
+_hd_launcher_tile_long_timeout (gpointer data)
+{
+  HdLauncherTile *tile = data;
+  HdLauncherTilePrivate *priv = tile->priv;
+
+  if (!priv->is_pressed)
+    return FALSE;
+
+  hd_launcher_tile_reset (tile);
+  g_signal_emit (tile, launcher_tile_signals[LONG_CLICKED], 0);
+
+  return FALSE;
+}
+
 static gboolean
 hd_launcher_tile_button_press (ClutterActor       *actor)
 {
@@ -494,6 +569,14 @@ hd_launcher_tile_button_press (ClutterActor       *actor)
   /* Set the 'pressed' flag */
   priv->is_pressed = TRUE;
 
+  if (priv->press_timeout)
+    {
+      g_source_remove (priv->press_timeout);
+      priv->press_timeout = 0;
+    }
+  priv->press_timeout = g_timeout_add (HD_LAUNCHER_TILE_LONG_PRESS_DUR,
+                                       _hd_launcher_tile_long_timeout,
+                                       actor);
   return TRUE;
 }
 
@@ -502,6 +585,12 @@ hd_launcher_tile_button_release (ClutterActor       *actor)
 {
   HdLauncherTilePrivate *priv = HD_LAUNCHER_TILE_GET_PRIVATE (actor);
 
+  if (priv->press_timeout)
+    {
+      g_source_remove (priv->press_timeout);
+      priv->press_timeout = 0;
+    }
+
   if (!priv->is_pressed)
     return TRUE;
 
@@ -517,6 +606,11 @@ hd_launcher_tile_dispose (GObject *gobject)
 {
   HdLauncherTilePrivate *priv = HD_LAUNCHER_TILE_GET_PRIVATE (gobject);
 
+  if (priv->press_timeout)
+    {
+      g_source_remove (priv->press_timeout);
+      priv->press_timeout = 0;
+    }
   if (priv->glow_timeline)
     {
       clutter_timeline_stop(priv->glow_timeline);
@@ -587,4 +681,9 @@ void hd_launcher_tile_reset(HdLauncherTile *tile)
   /* remove glow */
   hd_launcher_tile_set_glow(tile, FALSE);
   priv->is_pressed = FALSE;
+  if (priv->press_timeout)
+    {
+      g_source_remove (priv->press_timeout);
+      priv->press_timeout = 0;
+    }
 }
