@@ -93,10 +93,7 @@ hd_desktop_init (MBWMObject *this, va_list vap)
   geom.width  = HD_COMP_MGR_LANDSCAPE_WIDTH;
   geom.height = HD_COMP_MGR_LANDSCAPE_WIDTH;
 
-  hd_desktop_request_geometry (client, &geom,
-					 MBWMClientReqGeomForced);
-
-  mb_wm_client_focus (client);
+  hd_desktop_request_geometry (client, &geom, MBWMClientReqGeomForced);
 
   return 1;
 }
