@@ -20,6 +20,8 @@
 #include <string.h>
 #include <locale.h>
 
+#include "util/hd-transition.h"
+
 /* #define it something sane */
 #define TIDY_IS_SANE_BLUR_GROUP(obj)    ((obj) != NULL)
 
@@ -149,6 +151,9 @@ struct _TidyBlurGroupPrivate
 
   /* if anything changed we need to recalculate preblur */
   gboolean source_changed;
+
+  /* don't progress the animation for one clutter_actor_paint() */
+  gboolean skip_progress;
 };
 
 /**
@@ -183,18 +188,21 @@ gboolean tidy_blur_group_notify_modified_real(ClutterActor          *actor,
 
 static void tidy_blur_group_check_shader(TidyBlurGroup *group,
                                          ClutterShader **shader,
-                                         const char* fragment_source,
-                                         const char* vertex_source)
+                                         const char *fragment_source,
+                                         const char *vertex_source)
 {
   TidyBlurGroupPrivate *priv = group->priv;
+
   if (priv->use_shader && !*shader)
    {
-     GError           *error = NULL;
-     char             *old_locale;
+     GError *error = NULL;
+     char   *old_locale;
+
 #if GLSL_LOCALE_FIX
-      old_locale = g_strdup( setlocale (LC_ALL, NULL) );
+      old_locale = g_strdup (setlocale (LC_ALL, NULL));
       setlocale (LC_NUMERIC, "C");
 #endif
+
       *shader = clutter_shader_new();
       if (fragment_source)
         clutter_shader_set_fragment_source (*shader, fragment_source, -1);
@@ -216,22 +224,74 @@ static void tidy_blur_group_check_shader(TidyBlurGroup *group,
    }
 }
 
-static void tidy_blur_group_check_shaders(TidyBlurGroup *group)
+/* Allocate @priv->fbo_[ab]. */
+static void
+tidy_blur_group_allocate_textures (TidyBlurGroup *self)
 {
-  TidyBlurGroupPrivate *priv = group->priv;
-  tidy_blur_group_check_shader(group, &priv->shader_blur, BLUR_FRAGMENT_SHADER, BLUR_VERTEX_SHADER);
-  tidy_blur_group_check_shader(group, &priv->shader_saturate, SATURATE_FRAGMENT_SHADER, 0);
+  TidyBlurGroupPrivate *priv = self->priv;
+  guint tex_width, tex_height;
+
+#ifdef __i386__
+  if (!cogl_features_available(COGL_FEATURE_OFFSCREEN))
+    /* Don't try to allocate FBOs. */
+    return;
+#endif
+
+#if !RESIZE_TEXTURE
+  if (priv->fbo_a && priv->fbo_b)
+    /* Rotate in _paint() rather than resize. */
+    return;
+#endif
+
+  /* Free the textures. */
+  if (priv->fbo_a)
+    {
+      cogl_offscreen_unref(priv->fbo_a);
+      cogl_texture_unref(priv->tex_a);
+      priv->fbo_a = 0;
+      priv->tex_a = 0;
+    }
+  if (priv->fbo_b)
+    {
+      cogl_offscreen_unref(priv->fbo_b);
+      cogl_texture_unref(priv->tex_b);
+      priv->fbo_b = 0;
+      priv->tex_b = 0;
+    }
+
+  /* (Re)create the textures + offscreen buffers.  Downsample by 2.
+   * We can specify mipmapping here, but we don't need it. */
+  clutter_actor_get_size(CLUTTER_ACTOR(self), &tex_width, &tex_height);
+  tex_width  /= 2;
+  tex_height /= 2;
+
+  priv->tex_a = cogl_texture_new_with_size(
+            tex_width, tex_height, 0, FALSE /* mipmap */,
+            priv->use_alpha ? COGL_PIXEL_FORMAT_RGBA_8888 :
+                              COGL_PIXEL_FORMAT_RGB_565);
+  cogl_texture_set_filters(priv->tex_a, CGL_NEAREST, CGL_NEAREST);
+  priv->fbo_a = cogl_offscreen_new_to_texture(priv->tex_a);
+
+  priv->tex_b = cogl_texture_new_with_size(
+            tex_width, tex_height, 0, FALSE /* mipmap */,
+            priv->use_alpha ? COGL_PIXEL_FORMAT_RGBA_8888 :
+                              COGL_PIXEL_FORMAT_RGB_565);
+  cogl_texture_set_filters(priv->tex_b, CGL_NEAREST, CGL_NEAREST);
+  priv->fbo_b = cogl_offscreen_new_to_texture(priv->tex_b);
+
+  priv->current_blur_step = 0;
+  priv->source_changed = TRUE;
 }
 
 static gboolean
 tidy_blur_group_children_visible(ClutterGroup *group)
 {
-  gint n = clutter_group_get_n_children(group);
   gint i;
+  ClutterActor *actor;
 
-  for (i=0;i<n;i++)
+  for (i = 0, actor = clutter_group_get_nth_child(group, 0);
+       actor; actor = clutter_group_get_nth_child(group, ++i))
     {
-      ClutterActor *actor = clutter_group_get_nth_child(group, i);
       if (CLUTTER_IS_GROUP(actor))
         {
           if (tidy_blur_group_children_visible(CLUTTER_GROUP(actor)))
@@ -286,23 +346,14 @@ tidy_blur_group_fallback_blur(TidyBlurGroup *group, int tex_width, int tex_heigh
 
 /* If priv->chequer, draw a chequer pattern over the screen */
 static void
-tidy_blur_group_do_chequer(TidyBlurGroup *group)
+tidy_blur_group_do_chequer(TidyBlurGroup *group, guint width, guint height)
 {
-  gint            x_1, y_1, x_2, y_2;
-  gint            width, height;
-  ClutterColor    black = { 0x00, 0x00, 0x00, 0xFF };
-
+  static const ClutterColor black = { 0x00, 0x00, 0x00, 0xFF };
   TidyBlurGroupPrivate *priv = group->priv;
 
   if (!priv->chequer)
     return;
 
-  clutter_actor_get_allocation_coords (
-      CLUTTER_ACTOR(group), &x_1, &y_1, &x_2, &y_2);
-
-  width = x_2 - x_1;
-  height = y_2 - y_1;
-
   cogl_color (&black);
   cogl_texture_rectangle (priv->tex_chequer,
                           0, 0,
@@ -313,23 +364,69 @@ tidy_blur_group_do_chequer(TidyBlurGroup *group)
                           CFX_ONE*height/CHEQUER_SIZE);
 }
 
+/* Recursively set texture filtering state on this actor and children, and
+ * save the old state in the object. */
+static void
+recursive_set_linear_texture_filter(ClutterActor *actor, GArray *filters)
+{
+  if (CLUTTER_IS_CONTAINER(actor))
+    clutter_container_foreach(CLUTTER_CONTAINER(actor),
+                   (ClutterCallback)recursive_set_linear_texture_filter,
+                   filters);
+  else if (CLUTTER_IS_TEXTURE(actor))
+    {
+      ClutterTexture *tex = CLUTTER_TEXTURE(actor);
+      ClutterTextureQuality quality;
+
+      quality = clutter_texture_get_filter_quality(tex);
+      g_array_append_val(filters, quality);
+      clutter_texture_set_filter_quality(tex, GL_LINEAR);
+    }
+}
+
+/* Recursively set texture filtering state on this actor and children, and
+ * save the old state in the object. */
+static void
+recursive_reset_texture_filter(ClutterActor *actor,
+                               const ClutterTextureQuality **filtersp)
+{
+  if (CLUTTER_IS_CONTAINER(actor))
+    clutter_container_foreach(CLUTTER_CONTAINER(actor),
+                        (ClutterCallback)recursive_reset_texture_filter,
+                        filtersp);
+  else if (CLUTTER_IS_TEXTURE(actor))
+    {
+      clutter_texture_set_filter_quality(CLUTTER_TEXTURE(actor),
+                                         **filtersp);
+      (*filtersp)++;
+    }
+}
+
 /* An implementation for the ClutterGroup::paint() vfunc,
    painting all the child actors: */
 static void
 tidy_blur_group_paint (ClutterActor *actor)
 {
-  ClutterColor    white = { 0xff, 0xff, 0xff, 0xff };
-  ClutterColor    col = { 0xff, 0xff, 0xff, 0xff };
-  ClutterColor    bgcol = { 0x00, 0x00, 0x00, 0xff };
-  gint            x_1, y_1, x_2, y_2;
-  gint            steps_this_frame = 0;
+  static const ClutterColor white = { 0xff, 0xff, 0xff, 0xff };
+  static const ClutterColor bgcol = { 0x00, 0x00, 0x00, 0xff };
+  ClutterGroup *group         = CLUTTER_GROUP(actor);
+  TidyBlurGroup *container    = TIDY_BLUR_GROUP(group);
+  TidyBlurGroupPrivate *priv  = container->priv;
+  gint                         steps_this_frame = 0;
+  CoglHandle                   current_tex;
+  ClutterActorBox              box;
+  gint                         width, height, tex_width, tex_height;
+  gboolean                     rotate_90;
+  ClutterColor                 col;
+  GArray                      *filters;
+  const ClutterTextureQuality *filters_array;
 
   if (!TIDY_IS_SANE_BLUR_GROUP(actor))
     return;
 
-  ClutterGroup *group = CLUTTER_GROUP(actor);
-  TidyBlurGroup *container = TIDY_BLUR_GROUP(group);
-  TidyBlurGroupPrivate *priv = container->priv;
+  clutter_actor_get_allocation_box(actor, &box);
+  width  = CLUTTER_UNITS_TO_DEVICE(box.x2 - box.x1);
+  height = CLUTTER_UNITS_TO_DEVICE(box.y2 - box.y1);
 
   /* If we are rendering normally then shortcut all this, and
    just render directly without the texture */
@@ -340,95 +437,34 @@ tidy_blur_group_paint (ClutterActor *actor)
       priv->current_blur_step = 0;
       priv->source_changed = TRUE;
       /* render direct */
-      TIDY_BLUR_GROUP_GET_CLASS(actor)->overridden_paint(actor);
-      tidy_blur_group_do_chequer(container);
+      CLUTTER_ACTOR_CLASS(tidy_blur_group_parent_class)->paint(actor);
+      tidy_blur_group_do_chequer(container, width, height);
       return;
     }
 
-  tidy_blur_group_check_shaders(container);
-
-  clutter_actor_get_allocation_coords (actor, &x_1, &y_1, &x_2, &y_2);
-
 #ifdef __i386__
   if (!cogl_features_available(COGL_FEATURE_OFFSCREEN))
     { /* If we can't blur properly do something nicer instead :) */
       /* Otherwise crash... */
-      TIDY_BLUR_GROUP_GET_CLASS(actor)->overridden_paint(actor);
+      CLUTTER_ACTOR_CLASS(tidy_blur_group_parent_class)->paint(actor);
       col.blue = priv->brightness * 255;
       col.red = col.green = priv->brightness * 127;
       col.alpha = (1-priv->saturation) * 255;
       cogl_color (&col);
-      cogl_rectangle (0, 0, x_2 - x_1, y_2 - y_1);
-      tidy_blur_group_do_chequer(container);
+      cogl_rectangle (0, 0, width, height);
+      tidy_blur_group_do_chequer(container, width, height);
       return;
     }
 #endif
 
-  int width = x_2 - x_1;
-  int height = y_2 - y_1;
-  int exp_width = width/2;
-  int exp_height = height/2;
-  int tex_width = 0;
-  int tex_height = 0;
-  gboolean rotate_90;
-
-  /* check sizes */
-  if (priv->tex_a)
-    {
-      tex_width = cogl_texture_get_width(priv->tex_a);
-      tex_height = cogl_texture_get_height(priv->tex_a);
-    }
-
-#if RESIZE_TEXTURE
-  /* free texture if the size is wrong */
-  if (tex_width!=exp_width || tex_height!=exp_height) {
-    if (priv->fbo_a)
-      {
-        cogl_offscreen_unref(priv->fbo_a);
-        cogl_texture_unref(priv->tex_a);
-        priv->fbo_a = 0;
-        priv->tex_a = 0;
-      }
-    if (priv->fbo_b)
-      {
-        cogl_offscreen_unref(priv->fbo_b);
-        cogl_texture_unref(priv->tex_b);
-        priv->fbo_b = 0;
-        priv->tex_b = 0;
-      }
-    priv->current_blur_step = 0;
-    priv->source_changed = TRUE;
-  }
-#endif // RESIZE_TEXTURE
-  /* create the texture + offscreen buffer if they didn't exist.
-   * We can specify mipmapping here, but we don't need it */
-  if (!priv->tex_a)
-    {
-      tex_width = exp_width;
-      tex_height = exp_height;
-
-      priv->tex_a = cogl_texture_new_with_size(
-                tex_width, tex_height, 0, FALSE /*mipmap*/,
-                priv->use_alpha ? COGL_PIXEL_FORMAT_RGBA_8888 :
-                                  COGL_PIXEL_FORMAT_RGB_565);
-      cogl_texture_set_filters(priv->tex_a, CGL_NEAREST, CGL_NEAREST);
-      priv->fbo_a = cogl_offscreen_new_to_texture (priv->tex_a);
-    }
-  if (!priv->tex_b)
-    {
-      priv->tex_b = cogl_texture_new_with_size(
-                tex_width, tex_height, 0, 0,
-                priv->use_alpha ? COGL_PIXEL_FORMAT_RGBA_8888 :
-                                  COGL_PIXEL_FORMAT_RGB_565);
-      cogl_texture_set_filters(priv->tex_b, CGL_NEAREST, CGL_NEAREST);
-      priv->fbo_b = cogl_offscreen_new_to_texture (priv->tex_b);
-    }
+  tex_width  = cogl_texture_get_width(priv->tex_a);
+  tex_height = cogl_texture_get_height(priv->tex_a);
 
   /* It may be that we have resized, but the texture has not.
    * If so, try and keep blurring 'nice' by rotating so that
    * we don't have a texture that is totally the wrong aspect ratio */
-  rotate_90 = (tex_width > tex_height) != (width > height);
   /* If rotation has changed, trigger a redraw */
+  rotate_90 = (tex_width > tex_height) != (width > height);
   if (priv->current_is_rotated != rotate_90)
     {
       priv->current_is_rotated = rotate_90;
@@ -439,10 +475,8 @@ tidy_blur_group_paint (ClutterActor *actor)
   /* Draw children into an offscreen buffer */
   if (priv->source_changed && priv->current_blur_step==0)
     {
-      tidy_util_cogl_push_offscreen_buffer(priv->fbo_a);
       cogl_push_matrix();
-      /* translate a bit to let bilinear filter smooth out intermediate pixels */
-      cogl_translatex(CFX_ONE/2,CFX_ONE/2,0);
+      tidy_util_cogl_push_offscreen_buffer(priv->fbo_a);
 
       if (rotate_90) {
         cogl_scale(CFX_ONE*tex_width/height, CFX_ONE*tex_height/width);
@@ -453,13 +487,22 @@ tidy_blur_group_paint (ClutterActor *actor)
         cogl_scale(CFX_ONE*tex_width/width, CFX_ONE*tex_height/height);
       }
 
+      /* translate a bit to let bilinear filter smooth out intermediate pixels */
+      cogl_translatex(CFX_ONE/2,CFX_ONE/2,0);
 
       cogl_paint_init(&bgcol);
       cogl_color (&white);
-      TIDY_BLUR_GROUP_GET_CLASS(actor)->overridden_paint(actor);
+      /* Actually do the drawing of the children, but ensure that they are
+       * all linear sampled so they are smoothly interpolated. Restore after. */
+      filters = g_array_new(FALSE, FALSE, sizeof(ClutterTextureQuality));
+      recursive_set_linear_texture_filter(actor, filters);
+      CLUTTER_ACTOR_CLASS(tidy_blur_group_parent_class)->paint(actor);
+      filters_array = (void *)filters->data;
+      recursive_reset_texture_filter(actor, &filters_array);
+      g_array_free(filters, TRUE);
 
-      cogl_pop_matrix();
       tidy_util_cogl_pop_offscreen_buffer();
+      cogl_pop_matrix();
 
       priv->source_changed = FALSE;
       priv->current_blur_step = 0;
@@ -468,6 +511,9 @@ tidy_blur_group_paint (ClutterActor *actor)
       //g_debug("Rendered buffer");
       steps_this_frame++;
     }
+  else if (priv->skip_progress)
+    /* Progressing the animation doesn't play well with rotation. */
+    goto skip_progress;
 
   while (priv->current_blur_step < priv->blur_step &&
          steps_this_frame<MAX_STEPS_PER_FRAME)
@@ -515,6 +561,9 @@ tidy_blur_group_paint (ClutterActor *actor)
       priv->source_changed = TRUE;
     }
 
+skip_progress:
+  priv->skip_progress = FALSE;
+
   /* If we're still not blurred enough, ask to be rendered again... */
   if (priv->current_blur_step != priv->blur_step)
     clutter_actor_queue_redraw(actor);
@@ -554,7 +603,7 @@ tidy_blur_group_paint (ClutterActor *actor)
       cogl_clip_set(0, 0,
                     CLUTTER_INT_TO_FIXED(width),
                     CLUTTER_INT_TO_FIXED(height));
-      TIDY_BLUR_GROUP_GET_CLASS(actor)->overridden_paint(actor);
+      CLUTTER_ACTOR_CLASS(tidy_blur_group_parent_class)->paint(actor);
       cogl_clip_unset();
 
       /* If we're zooming less than 1, we want to re-render everything
@@ -582,7 +631,7 @@ tidy_blur_group_paint (ClutterActor *actor)
                   cogl_clip_set(0, 0,
                                 CLUTTER_INT_TO_FIXED(width),
                                 CLUTTER_INT_TO_FIXED(height));
-                  TIDY_BLUR_GROUP_GET_CLASS(actor)->overridden_paint(actor);
+                  CLUTTER_ACTOR_CLASS(tidy_blur_group_parent_class)->paint(actor);
                   cogl_clip_unset();
                   cogl_pop_matrix();
                 }
@@ -596,7 +645,7 @@ tidy_blur_group_paint (ClutterActor *actor)
 
   if (col.alpha == 0)
     {
-      tidy_blur_group_do_chequer(container);
+      tidy_blur_group_do_chequer(container, width, height);
       return;
     }
 
@@ -620,9 +669,14 @@ tidy_blur_group_paint (ClutterActor *actor)
       cogl_translatex(-CFX_ONE*width/2, -CFX_ONE*height/2, 0);
     }
 
+  /* Set the blur texture to linear interpolation - so we draw it smoothly
+   * Onto the screen */
+  current_tex = priv->current_is_a ? priv->tex_a : priv->tex_b;
+  cogl_texture_set_filters(current_tex, CGL_LINEAR, CGL_LINEAR);
+
   if ((priv->zoom >= 1) || !priv->use_mirror)
     {
-      cogl_texture_rectangle (priv->current_is_a ? priv->tex_a : priv->tex_b,
+      cogl_texture_rectangle (current_tex,
                               mx-zx, my-zy,
                               mx+zx, my+zy,
                               0, 0, CFX_ONE, CFX_ONE);
@@ -728,12 +782,15 @@ tidy_blur_group_paint (ClutterActor *actor)
             v+=6;
           }
       /* render! */
-      cogl_texture_triangles (priv->current_is_a ? priv->tex_a : priv->tex_b,
+      cogl_texture_triangles (current_tex,
                               6*(VIGNETTE_TILES*VIGNETTE_TILES),
                               verts,
                               TRUE);
     }
 
+  /* Reset the filters on the current texture ready for normal blurring */
+  cogl_texture_set_filters(current_tex, CGL_NEAREST, CGL_NEAREST);
+
   if (rotate_90)
     {
       cogl_pop_matrix();
@@ -742,7 +799,7 @@ tidy_blur_group_paint (ClutterActor *actor)
   if (priv->use_shader && priv->shader_saturate)
     clutter_shader_set_is_enabled (priv->shader_saturate, FALSE);
 
-  tidy_blur_group_do_chequer(container);
+  tidy_blur_group_do_chequer(container, width, height);
 }
 
 static void
@@ -785,8 +842,8 @@ tidy_blur_group_class_init (TidyBlurGroupClass *klass)
   gobject_class->dispose = tidy_blur_group_dispose;
 
   /* Provide implementations for ClutterActor vfuncs: */
-  klass->overridden_paint = actor_class->paint;
-  actor_class->paint = tidy_blur_group_paint;
+  if (!hd_transition_get_int("blur", "turbo", 0))
+    actor_class->paint = tidy_blur_group_paint;
   actor_class->notify_modified = tidy_blur_group_notify_modified_real;
 }
 
@@ -851,6 +908,14 @@ tidy_blur_group_init (TidyBlurGroup *self)
       COGL_PIXEL_FORMAT_A_8,
       CHEQUER_SIZE,
       dither_data);
+
+  tidy_blur_group_check_shader(self, &priv->shader_blur,
+                               BLUR_FRAGMENT_SHADER, BLUR_VERTEX_SHADER);
+  tidy_blur_group_check_shader(self, &priv->shader_saturate,
+                               SATURATE_FRAGMENT_SHADER, 0);
+
+  g_signal_connect(self, "notify::allocation",
+                   G_CALLBACK(tidy_blur_group_allocate_textures), NULL);
 }
 
 /*
@@ -1081,6 +1146,16 @@ void tidy_blur_group_hint_source_changed(ClutterActor *blur_group)
   priv->source_changed = TRUE;
 }
 
+void tidy_blur_group_stop_progressing(ClutterActor *blur_group)
+{
+  TidyBlurGroupPrivate *priv;
+
+  if (!TIDY_IS_SANE_BLUR_GROUP(blur_group))
+    return;
+
+  priv = TIDY_BLUR_GROUP(blur_group)->priv;
+  priv->skip_progress = TRUE;
+}
 
 /**
  * tidy_blur_group_source_buffered:
