@@ -49,7 +49,8 @@
 #include "hd-util.h"
 
 /* The master of puppets */
-#define TRANSITIONS_INI "/usr/share/hildon-desktop/transitions.ini"
+#define TRANSITIONS_INI             "/usr/share/hildon-desktop/transitions.ini"
+#define TRANSITIONS_INI_FROM_THEME  "/etc/hildon/theme/transitions.ini"
 
 typedef struct _HDEffectData
 {
@@ -61,6 +62,8 @@ typedef struct _HDEffectData
   MBWMCompMgrClutterClient *cclient2;
   ClutterActor             *cclient2_actor;
   HdCompMgr                *hmgr;
+  /* hd_render_manager_set_visibilities() when all transitions are done */
+  gboolean                  fixup_visibilities;
   /* original/expected position of application/menu */
   ClutterGeometry           geo;
   /* used in rotate_screen to set the direction (and amount) of movement */
@@ -108,12 +111,30 @@ static struct
 
   /*
    * What is *_fsm() currently doing:
-   * -- #IDLE:      nothing, we're sitting in landscape or portrait
-   * -- #TRANS_START: we're hoping to be called back on idle to start the
-   *                  animation
-   * -- #FADE_OUT:  hd_transition_fade_and_rotate() is fading out
-   * -- #WAITING:   for X to finish reconfiguring the screen
-   * -- #FADE_IN:   second hd_transition_fade_and_rotate() is in progress
+   * -- #IDLE:
+   *    Nothing, we're sitting in landscape or portrait.
+   *    Freeze the display and start reconfiguring the windows.
+   * -- #TRANS_START:
+   *    Windows reconfigured, start fading.
+   * -- #FADE_OUT:
+   *    Faded out to blankness, reconfigure the screen and wait until
+   *    X is done.
+   * -- #WAIT_FOR_ROOT_CONFIG:
+   *    The root window is reconfigured, the screen is now officially
+   *    in portrait.  Wait for a while for the initial application
+   *    updates until thawing the display.
+   * -- #WAIT_FOR_DAMAGES:
+   *    Damage done, start fading in.
+   * -- #FADE_IN:
+   *    Mission complete.
+   * -- #RECOVER:
+   *    This is a special phase, not reached normally.  We only get here
+   *    if the direction changed during the transition to IDLE->TRANS_START.
+   *    Then we proceed to #FADE_IN.
+   *
+   * The FSM tries to do the right thing when it's instructed to change
+   * direction in the middle of the transition, so the phases may jump
+   * from here to there.
    */
   enum
   {
@@ -122,14 +143,16 @@ static struct
     FADE_OUT,
     WAIT_FOR_ROOT_CONFIG,
     WAIT_FOR_DAMAGES,
+    RECOVER,
     FADE_IN,
   } phase;
 
   /*
    * @goto_state when we've %FADE_OUT:d.  Set by
    * hd_transition_rotate_screen_and_change_state()
-   * Its initial value is %HDRM_STATE_UNDEFINED, which means don't
-   * change the state. */
+   * Its initial value is %HDRM_STATE_UNDEFINED,
+   * which means don't change the state.
+   */
   HDRMStateEnum goto_state;
 
   gulong root_config_signal_id;
@@ -142,8 +165,22 @@ static struct
   /* This timer counts from when we first entered the WAITING state,
    * so if we are continually getting damage we don't just hang there. */
   GTimer *timer;
+
+  /*
+   * The number of %_MAEMO_ROTATION_PATIENCE requests we received since
+   * we notified our clients about the screen size change.  We we get one
+   * we extend our patience with regards to client redraws to our maximum.
+   * If we were asked for patience and then notified that it's not longer
+   * necessary we stop waiting for damages immedeately.
+   */
+  guint patience_requests;
 } Orientation_change;
 
+/* The number of transitions in progress requesting for @fixup_visibilities.
+ * At the moment only the popup (menus and dialogs), the fade (notes, banners)
+ * and subview transitions are involved. */
+static guint Transitions_running;
+
 /* If %TRUE keep reloading transitions.ini until we can
  * and we can watch it. */
 static gboolean transitions_ini_is_dirty;
@@ -357,7 +394,7 @@ on_popup_timeline_new_frame(ClutterTimeline *timeline,
    * a habit of changing size while they move. */
   clutter_actor_get_geometry(actor, &geo);
 
-  pop_top = geo.y==0;
+  pop_top = geo.y<=0;
   pop_bottom = geo.y+geo.height==hd_comp_mgr_get_current_screen_height();
   if (pop_top && pop_bottom)
     pop_top = FALSE;
@@ -370,7 +407,7 @@ on_popup_timeline_new_frame(ClutterTimeline *timeline,
 
   if (pop_top)
     {
-      status_low = -geo.height;
+      status_low = geo.y-geo.height;
       status_high = geo.y;
     }
   else if (pop_bottom)
@@ -471,7 +508,7 @@ on_close_timeline_new_frame(ClutterTimeline *timeline,
   if (!CLUTTER_IS_ACTOR(actor))
     return;
 
-  if (hd_dbus_display_is_off && data)
+  if (hd_dbus_display_is_off)
     {
       for (i = 0; i < HDCM_UNMAP_PARTICLES; ++i)
         if (data->particles[i])
@@ -548,19 +585,23 @@ on_notification_timeline_new_frame(ClutterTimeline *timeline,
   float now;
   ClutterActor *actor;
   guint width, height;
+  gint tbw, px, py;
 
   actor = data->cclient_actor;
   if (!CLUTTER_IS_ACTOR(actor) || hd_dbus_display_is_off)
     return;
 
+  tbw = hd_title_bar_get_button_width( /* Task Button's current width. */
+                        HD_TITLE_BAR(hd_render_manager_get_title_bar()));
   clutter_actor_get_size(actor, &width, &height);
-
+  clutter_actor_get_position(actor, &px, &py);
   now = frame_num / (float)clutter_timeline_get_n_frames(timeline);
 
-  if (hd_comp_mgr_is_portrait())
+  if (hd_comp_mgr_is_portrait()
+      && hd_transition_get_int("notification", "is_cool", 0))
     {
-      /* In portrait swipe from right to left, stay in the corner
-       * then swipe out, following a bezier curve.  At the start
+      /* In portrait fly from right to left, stay in the corner
+       * then fly away, following a bezier curve.  At the start
        * the notification's bottom-center is at the screen's
        * top-right.  By the end the notification's bottom-right
        * is at the screen's top-left. */
@@ -608,31 +649,41 @@ on_notification_timeline_new_frame(ClutterTimeline *timeline,
 
       if (now < thr)
         { /* fade, move, resize */
+          float sc;
           gint cx, cy;
-          float sx, sy;
+          gint dx, dy, dw, dh;
 
           /*
-           * visual geometry: 366x88+112+0 -> 96x23+8+17
+           * visual geometry: 366x88+112+0 -> 48x11+32+22 or
+           *                  366x88+80+0  -> 16x3+32+26
            *                  scale it down proportionally
            *                  and place it in the middle of the tasks button
-           *                  leaving 8 pixels left and right
-           * opacity:         1 -> 0.75
+           *                  leaving 8 pixels left and right, keeping the
+           *                  aspect ratio
+           * opacity:         1 -> 0.50
            * use smooth ramping
            */
-          t = hd_transition_smooth_ramp(now / thr);
-          cx = ( 8 - clutter_actor_get_x(actor))*t;
-          cy = (17 - clutter_actor_get_y(actor))*t;
-          sx = ((96.0/width  - 1)*t + 1);
-          sy = ((23.0/height - 1)*t + 1);
 
-          clutter_actor_set_scale (actor, sx, sy);
-          clutter_actor_set_anchor_point (actor, -cx/sx, -cy/sy);
-          clutter_actor_set_opacity(actor, 255 * ((0.75 - 1)*t + 1));
+          /* It's probably best to count it with your fingers to follow
+           * this mumble-jumbo. */
+          dx = 32;
+          dw = tbw - 2*dx;
+          dh = (float)dw/width * height;
+          dy = (HD_COMP_MGR_TOP_LEFT_BTN_HEIGHT - dh) / 2;
+
+          t = hd_transition_smooth_ramp(now / thr);
+          cx = (dx - tbw)*t + tbw - px;
+          cy = (dy -  py)*t;
+          sc = (((float)dw/width  - 1)*t + 1);
+
+          clutter_actor_set_scale (actor, sc, sc);
+          clutter_actor_set_anchor_point (actor, -cx/sc, -cy/sc);
+          clutter_actor_set_opacity(actor, 255 * ((0.50 - 1)*t + 1));
         }
       else
         { /* fade: 0.75 -> 0 linearly */
           t = (now - thr) / (1.0 - thr);
-          clutter_actor_set_opacity(actor, 255 * (-0.75*t + 0.75));
+          clutter_actor_set_opacity(actor, 255 * (-0.50*t + 0.50));
         }
     }
   else
@@ -643,7 +694,8 @@ on_notification_timeline_new_frame(ClutterTimeline *timeline,
       float scale =  1 + (1-amt)*0.5f;
       float ang = amt * 3.141592f * 0.5f;
       float corner_x = (hd_comp_mgr_get_current_screen_width()*0.5f
-                        - HD_COMP_MGR_TOP_LEFT_BTN_WIDTH) * cos(ang);
+                        - HD_COMP_MGR_TOP_LEFT_BTN_WIDTH) * cos(ang)
+                       - px + tbw;
       float corner_y = (sin(ang)-1) * height;
       /* We set anchor point so if the notification resizes/positions
        * in flight, we're ok.  NOTE that the position of the actor
@@ -840,7 +892,20 @@ hd_transition_completed (ClutterTimeline* timeline, HDEffectData *data)
                                         G_CALLBACK (on_screen_size_changed),
                                         data);
 
+  /* @Transitions_running only accounts for transitions asking for
+   * @fixup_visibilities.  If we're finishing off the last one it
+   * must be safe (knock-knock-knock) to re-evaluate visibilities. */
+  if (data->fixup_visibilities && --Transitions_running == 0
+      && STATE_IS_APP(hd_render_manager_get_state()))
+    {
+      hd_render_manager_set_visibilities();
+      hd_comp_mgr_portrait_or_not_portrait(MB_WM_COMP_MGR(hmgr), NULL);
+    }
+
   g_free (data);
+
+  if (hmgr)
+    hd_comp_mgr_reconsider_compositing (MB_WM_COMP_MGR (hmgr));
 }
 
 void
@@ -875,6 +940,7 @@ hd_transition_popup(HdCompMgr                  *mgr,
   g_signal_connect (data->timeline, "completed",
                         G_CALLBACK (hd_transition_completed), data);
   data->geo = geo;
+  Transitions_running += data->fixup_visibilities = TRUE;
 
   /*
    * If @actor is a fullscreen dialog it's not in apptop but in home_blur.
@@ -896,6 +962,7 @@ hd_transition_popup(HdCompMgr                  *mgr,
 
   /* Add actor for the background when we pop a bit too far */
   data->particles[0] = g_object_ref(clutter_rectangle_new());
+  clutter_actor_set_name(data->particles[0], "popup background");
   clutter_container_add_actor(CLUTTER_CONTAINER(actor), data->particles[0]);
   hd_gtk_style_get_bg_color(HD_GTK_BUTTON_SINGLETON, GTK_STATE_NORMAL,
                               &col);
@@ -925,6 +992,7 @@ hd_transition_fade(HdCompMgr                  *mgr,
       mb_wm_comp_mgr_clutter_client_get_actor( data->cclient ) );
   data->hmgr = HD_COMP_MGR (mgr);
   data->timeline = hd_transition_timeline_new("fade", event, 250);
+  Transitions_running += data->fixup_visibilities = TRUE;
 
   if (HD_IS_BANNER_NOTE(c))
     {
@@ -1128,6 +1196,10 @@ hd_transition_notification(HdCompMgr                  *mgr,
 
   /* first call to stop flicker */
   on_notification_timeline_new_frame(data->timeline, 0, data);
+  /* Show the actor and add it to the front group */
+  clutter_actor_show(data->cclient_actor);
+  hd_render_manager_add_to_front_group(data->cclient_actor);
+  /* Finally start the timeline... */
   clutter_timeline_start (data->timeline);
 }
 
@@ -1246,6 +1318,7 @@ hd_transition_subview(HdCompMgr                  *mgr,
   data->cclient2_actor = g_object_ref (
       mb_wm_comp_mgr_clutter_client_get_actor( data->cclient2 ) );
   data->hmgr = HD_COMP_MGR (mgr);
+  Transitions_running += data->fixup_visibilities = TRUE;
   data->timeline = hd_transition_timeline_new("subview", event, 250);
 
   g_signal_connect (data->timeline, "new-frame",
@@ -1323,6 +1396,7 @@ hd_transition_fade_and_rotate(gboolean first_part,
     data->angle *= -1;
   /* Add the actor we use to dim out the screen */
   data->particles[0] = g_object_ref(clutter_rectangle_new_with_color(&black));
+  clutter_actor_set_name(data->particles[0], "black rotation background");
   clutter_actor_set_size(data->particles[0],
       hd_comp_mgr_get_current_screen_width (),
       hd_comp_mgr_get_current_screen_height ());
@@ -1336,6 +1410,7 @@ hd_transition_fade_and_rotate(gboolean first_part,
        * portrait half of the animation. This is pretty nasty, but as the home
        * applets aren't repositioned they can sometimes be seen in the background.*/
        data->particles[1] = g_object_ref(clutter_rectangle_new_with_color(&black));
+       clutter_actor_set_name(data->particles[1], "other rotation background");
        clutter_actor_set_position(data->particles[1],
            HD_COMP_MGR_LANDSCAPE_HEIGHT, 0);
        clutter_actor_set_size(data->particles[1],
@@ -1352,10 +1427,39 @@ hd_transition_fade_and_rotate(gboolean first_part,
   clutter_timeline_start (data->timeline);
 }
 
+/* Process %_MAEMO_ROTATION_PATIENCE requests. */
 static void
-hd_transition_damage_timer_destroyed (gpointer unused)
+patience (XClientMessageEvent *event, void *unused)
 {
-  Orientation_change.timeout_id = NULL;
+  /* Is somebody probing our patience? */
+  if (event->message_type != hd_comp_mgr_get_atom (
+                          HD_COMP_MGR (Orientation_change.wm->comp_mgr),
+                          HD_ATOM_MAEMO_ROTATION_PATIENCE))
+    return;
+
+  if (event->data.l[0])
+    { /* Grant @max:imal patience. */
+      gint max;
+
+      if (Orientation_change.timeout_id)
+        {
+          /* remaining := max(damage_timeout_max-elapsed, 0) */
+          max  = hd_transition_get_int("rotate", "damage_timeout_max", 1000);
+          max -= g_timer_elapsed(Orientation_change.timer, NULL) * 1000.0;
+          if (max > 0)
+            Orientation_change.timeout_id->remaining = max;
+        }
+      if (Orientation_change.phase <= WAIT_FOR_DAMAGES)
+        Orientation_change.patience_requests++;
+    }
+  else
+    { /* Get out of WAIT_FOR_DAMAGES as quickly as possible. */
+      if (Orientation_change.patience_requests)
+        Orientation_change.patience_requests--;
+      if (!Orientation_change.patience_requests
+          && Orientation_change.timeout_id)
+        Orientation_change.timeout_id->remaining = 0;
+    }
 }
 
 static gboolean
@@ -1379,19 +1483,36 @@ hd_transition_rotating_fsm(void)
           CLUTTER_ACTOR(hd_render_manager_get()), 1);
         tidy_cached_group_set_downsampling_factor(
           CLUTTER_ACTOR(hd_render_manager_get()), 1);
-        /* Super massive extra large hack to remove status area when launching
-         * phone app from the launcher */
+        /* Stop displaying the loading screenshot, which was displayed
+         * as a small square just over the icon when launching phone.
+         * However, leave it alone if it's already there fully grown. */
+        if (hd_launcher_transition_is_playing())
+          hd_render_manager_set_loading(NULL);
+        /* Super massive extra large hack to remove status area when
+         * launching phone app from the launcher */
         if (hd_render_manager_get_previous_state()==HDRM_STATE_LAUNCHER)
-          clutter_actor_hide(hd_render_manager_get_status_area());
+          {
+            ClutterActor *hsm = hd_render_manager_get_status_area();
+            if (hsm) clutter_actor_hide(hsm);
+          }
+        else if (STATE_IS_APP(hd_render_manager_get_previous_state()) &&
+                 STATE_IS_APP(hd_render_manager_get_state()))
+          /* Less super hack to not show wrong title while rotating - eg from
+           * calculator to phone. We must only do it in this case, as if going
+           * phone->desktop we will kill the title bar too soon. Or if going
+           * phone->task_nav (which sets state to APP first) the buttons are
+           * the wrong size. */
+          hd_title_bar_update_now(HD_TITLE_BAR(hd_render_manager_get_title_bar()));
         /* Force redraw for screenshot *now*, before windows have a
-         * chance to change */
+         * chance to change.  Tell the render manager not to progress
+         * the animation, it will be reset anyway. */
+        hd_render_manager_pause_blur_animation();
         clutter_redraw(CLUTTER_STAGE(clutter_stage_get_default()));
         /* Start rotate transition */
         hd_util_set_rotating_property(Orientation_change.wm, TRUE);
         /* Layout windows at the rotated size */
-        hd_util_set_screen_size_properties(Orientation_change.wm,
-            Orientation_change.direction == GOTO_PORTRAIT ? 480 : 800,
-            Orientation_change.direction == GOTO_PORTRAIT ? 800 : 480);
+        hd_util_set_screen_size_property(Orientation_change.wm,
+                         Orientation_change.direction == GOTO_PORTRAIT);
         Orientation_change.wm->flags |= MBWindowManagerFlagLayoutRotated;
         mb_wm_layout_update(Orientation_change.wm->layout);
         /* We now call ourselves back on idle. The idea is that the sudden
@@ -1401,12 +1522,20 @@ hd_transition_rotating_fsm(void)
         g_idle_add((GSourceFunc)(hd_transition_rotating_fsm), NULL);
         break;
       case TRANS_START:
-        /* Fade to black ((c) Metallica) */
-        Orientation_change.phase = FADE_OUT;
-        hd_transition_fade_and_rotate(
-                        TRUE, Orientation_change.direction == GOTO_PORTRAIT,
-                        G_CALLBACK(hd_transition_rotating_fsm), NULL);
-        break;
+        if (Orientation_change.direction == Orientation_change.new_direction)
+          {
+            /* Fade to black ((c) Metallica) */
+            Orientation_change.phase = FADE_OUT;
+            hd_transition_fade_and_rotate(
+                            TRUE, Orientation_change.direction == GOTO_PORTRAIT,
+                            G_CALLBACK(hd_transition_rotating_fsm), NULL);
+            break;
+          }
+        else
+          {
+            Orientation_change.direction = Orientation_change.new_direction;
+            goto trans_start_error;
+          }
       case FADE_OUT:
         /*
          * We're faded out, now it is time to change HDRM state
@@ -1469,6 +1598,7 @@ hd_transition_rotating_fsm(void)
         /* Fall through */
       case WAIT_FOR_ROOT_CONFIG:
       case WAIT_FOR_DAMAGES:
+trans_start_error:
         if (Orientation_change.direction != Orientation_change.new_direction)
           {
             if (Orientation_change.root_config_signal_id)
@@ -1494,33 +1624,60 @@ hd_transition_rotating_fsm(void)
 
             g_assert(!Orientation_change.timeout_id);
             Orientation_change.timeout_id = hptimer_new(
-                  hd_transition_get_int("rotate", "damage_timeout", 50),
-                  (GSourceFunc)hd_transition_rotating_fsm, NULL,
-                  hd_transition_damage_timer_destroyed);
+                  Orientation_change.patience_requests
+                    ? hd_transition_get_int("rotate", "damage_timeout_max",
+                                            1000)
+                    : hd_transition_get_int("rotate", "damage_timeout", 50),
+                  (GSourceFunc)hd_transition_rotating_fsm,
+                  &Orientation_change.timeout_id,
+                  (GDestroyNotify)g_nullify_pointer);
+            g_timer_start(Orientation_change.timer);
             Orientation_change.phase = WAIT_FOR_DAMAGES;
           }
-        else /* WAIT_FOR_DAMAGES || FADE_OUT error path */
-          { /* Fade back in */
+        else /* WAIT_FOR_DAMAGES || FADE_OUT error path || TRANS_START error */
+          {
             /* We must update the layout again so the window sizes
              * return to normal relative to the screen. flags is probably
              * already correct. But just for safety. */
-            hd_util_set_screen_size_properties(Orientation_change.wm,
-                        Orientation_change.direction == GOTO_PORTRAIT ? 480 : 800,
-                        Orientation_change.direction == GOTO_PORTRAIT ? 800 : 480);
+            hd_util_set_screen_size_property(Orientation_change.wm,
+                         Orientation_change.direction == GOTO_PORTRAIT);
             Orientation_change.wm->flags &= ~MBWindowManagerFlagLayoutRotated;
             mb_wm_layout_update(Orientation_change.wm->layout);
-            /* Undo the redraw stopping that happened in FADE_OUT */
-            Orientation_change.phase = FADE_IN;
-            clutter_actor_set_allow_redraw(
-                            CLUTTER_ACTOR(hd_render_manager_get()), TRUE);
-            clutter_actor_show(CLUTTER_ACTOR(hd_render_manager_get()));
-            hd_transition_fade_and_rotate(
-                    FALSE, Orientation_change.direction == GOTO_PORTRAIT,
-                    G_CALLBACK(hd_transition_rotating_fsm), NULL);
-            /* Fix NB#117109 by re-evaluating what is blurred and what isn't */
-            hd_render_manager_restack();
+            if (Orientation_change.phase > TRANS_START)
+              { /* Fade back in */
+                /* Undo the redraw stopping that happened in FADE_OUT */
+                Orientation_change.phase = FADE_IN;
+                clutter_actor_set_allow_redraw(
+                                CLUTTER_ACTOR(hd_render_manager_get()), TRUE);
+                clutter_actor_show(CLUTTER_ACTOR(hd_render_manager_get()));
+                hd_transition_fade_and_rotate(
+                        FALSE, Orientation_change.direction == GOTO_PORTRAIT,
+                        G_CALLBACK(hd_transition_rotating_fsm), NULL);
+                /* Fix NB#117109 by re-evaluating what is blurred and what isn't */
+                hd_render_manager_restack();
+              }
+            else
+              {
+                /*
+                 * Direction changed in TRANS_START.  Half of what IDLE did
+                 * is already undone, but the blur group is still frozen.
+                 * Let's wait a bit until the windows are reconfigured again,
+                 * then toast it.
+                 */
+                Orientation_change.phase = RECOVER;
+                g_idle_add((GSourceFunc)(hd_transition_rotating_fsm), NULL);
+              }
           }
         break;
+      case RECOVER:
+        /* Thaw the blur group.  Only from the TRANS_START error path can we
+         * get here. */
+        tidy_cached_group_changed(CLUTTER_ACTOR(hd_render_manager_get()));
+        tidy_cached_group_set_render_cache(
+                                  CLUTTER_ACTOR(hd_render_manager_get()), 0);
+        tidy_cached_group_set_downsampling_factor(
+                                  CLUTTER_ACTOR(hd_render_manager_get()), 0);
+        /* Fall through */
       case FADE_IN:
         {
           ClutterActor *actor;
@@ -1536,7 +1693,12 @@ hd_transition_rotating_fsm(void)
             /* No sense resetting the rotating property. */
             hd_transition_rotating_fsm();
           else
-            hd_util_set_rotating_property(Orientation_change.wm, FALSE);
+            { /* We're finally settled. */
+              hd_comp_mgr_reconsider_compositing (
+                            Orientation_change.wm->comp_mgr);
+              hd_util_set_rotating_property (Orientation_change.wm, FALSE);
+              Orientation_change.patience_requests = 0;
+            }
           break;
         }
     }
@@ -1550,19 +1712,27 @@ hd_transition_rotating_fsm(void)
 gboolean
 hd_transition_rotate_screen (MBWindowManager *wm, gboolean goto_portrait)
 { g_debug("%s(goto_portrait=%d)", __FUNCTION__, goto_portrait);
+  static unsigned long cmsg_id;
+
+  /* Initialize the fsm infrastructure. */
   Orientation_change.wm = wm;
   if (!Orientation_change.timer)
     Orientation_change.timer = g_timer_new();
+  if (!cmsg_id)
+    cmsg_id = mb_wm_main_context_x_event_handler_add (wm->main_ctx,
+                                               wm->root_win->xwindow,
+                                               ClientMessage,
+                                               (MBWMXEventFunc)patience,
+                                               NULL);
 
   Orientation_change.new_direction = goto_portrait
     ? GOTO_PORTRAIT : GOTO_LANDSCAPE;
-
   if (Orientation_change.phase == IDLE)
     {
       if (goto_portrait == hd_comp_mgr_is_portrait ())
         {
-          g_warning("%s: already in %s mode", __FUNCTION__,
-                    goto_portrait ? "portrait" : "landscape");
+          g_debug("%s: already in %s mode", __FUNCTION__,
+                  goto_portrait ? "portrait" : "landscape");
           return FALSE;
         }
 
@@ -1586,6 +1756,17 @@ hd_transition_rotate_screen_and_change_state (HDRMStateEnum state)
   Orientation_change.goto_state = state;
 }
 
+/* Returns whether the rotating fsm is going to change hdrm state.
+ * In theory it may be wrong but in practice it's only used for
+ * APP->TASK_NAV and hdrm make sure @goto_state is cleared otherwise. */
+gboolean
+hd_transition_rotation_will_change_state (void)
+{
+  return (Orientation_change.phase == TRANS_START
+          || Orientation_change.phase == FADE_OUT)
+    && Orientation_change.goto_state != HDRM_STATE_UNDEFINED;
+}
+
 /* Are we in the middle of a change? */
 gboolean
 hd_transition_is_rotating (void)
@@ -1614,14 +1795,30 @@ hd_transition_rotate_ignore_damage()
     {
       gint max;
 
-      /* Only postpone the timeout if we haven't postponed
+      /*
+       * Only postpone the timeout if we haven't postponed
        * it too long already. This stops us getting stuck
-       * in the WAITING state if an app keeps redrawing. */
-      max = hd_transition_get_int("rotate", "damage_timeout_max", 1000);
-      Orientation_change.timeout_id->expiry =
-          g_timer_elapsed(Orientation_change.timer, NULL) < max / 1000.0
-            ? hd_transition_get_int("rotate", "damage_timeout_plus", 50)
-            : 0;
+       * in the WAITING state if an app keeps redrawing.
+       *
+       * remaining := min(max(remaining, damage_timeout_plus),
+       *                  max(damage_timeout_max-elapsed, 0))
+       */
+      max  = hd_transition_get_int("rotate", "damage_timeout_max", 1000);
+      max -= g_timer_elapsed(Orientation_change.timer, NULL) * 1000.0;
+      if (max > 0)
+        {
+          gint remaining;
+
+          remaining = hd_transition_get_int("rotate", "damage_timeout_plus",
+                                            50);
+          if (Orientation_change.timeout_id->remaining < remaining)
+            Orientation_change.timeout_id->remaining = remaining;
+          if (Orientation_change.timeout_id->remaining > max)
+            Orientation_change.timeout_id->remaining = max;
+        }
+      else
+        Orientation_change.timeout_id->remaining = 0;
+
       return TRUE;
     }
   return FALSE;
@@ -1638,9 +1835,13 @@ hd_transition_actor_will_go_away (ClutterActor *actor)
   if (!(hcmgrc = g_object_get_data(G_OBJECT(actor),
                                    "HD-MBWMCompMgrClutterClient")))
     return FALSE;
-  if (!hcmgrc->effect || hcmgrc->effect->event != MBWMCompMgrClientEventUnmap)
+  if (!hcmgrc->effect)
     return FALSE;
-  return hcmgrc->effect->cclient == MB_WM_COMP_MGR_CLUTTER_CLIENT (hcmgrc);
+  if (hcmgrc->effect->event == MBWMCompMgrClientEventMap)
+    return hcmgrc->effect->cclient2 == MB_WM_COMP_MGR_CLUTTER_CLIENT (hcmgrc);
+  if (hcmgrc->effect->event == MBWMCompMgrClientEventUnmap)
+    return hcmgrc->effect->cclient  == MB_WM_COMP_MGR_CLUTTER_CLIENT (hcmgrc);
+  return FALSE;
 }
 
 /* Start playing @fname asynchronously. */
@@ -1696,6 +1897,14 @@ hd_transition_play_sound (const gchar * fname)
 
 }
 
+/* We want to call this when the theme changes, as transitions.ini *could*
+ * be loaded from the theme. */
+void
+hd_transition_set_file_changed(void) {
+  g_debug("%s: setting transitions.ini modified", __FUNCTION__);
+  transitions_ini_is_dirty = 2*TRUE;
+}
+
 static gboolean
 transitions_ini_changed(GIOChannel *chnl, GIOCondition cond, gpointer unused)
 {
@@ -1705,20 +1914,17 @@ transitions_ini_changed(GIOChannel *chnl, GIOCondition cond, gpointer unused)
       != G_IO_STATUS_NORMAL)
     {
       g_warning("%s: g_io_channel_read_chars failed", __func__);
-      g_io_channel_unref(chnl);
-      return FALSE;
     }
-  if (ibuf.mask & (IN_MODIFY | IN_DELETE_SELF | IN_MOVE_SELF | IN_IGNORED))
+  else if (ibuf.mask & (IN_MODIFY | IN_DELETE_SELF | IN_MOVE_SELF | IN_IGNORED))
     {
       g_debug("disposing transitions.ini");
-      transitions_ini_is_dirty = TRUE;
+      transitions_ini_is_dirty++;
 
       /* Track no more if the dirent changed or disappeared. */
       if (ibuf.mask & (IN_MOVE_SELF | IN_DELETE_SELF | IN_IGNORED))
         {
           g_debug("watching no more");
-          g_io_channel_unref(chnl);
-          return FALSE;
+          transitions_ini_is_dirty++;
         }
     }
   return TRUE;
@@ -1731,54 +1937,69 @@ hd_transition_get_keyfile(void)
   static GIOChannel *transitions_ini_watcher;
   GError *error;
   GKeyFile *ini;
+  const char *fname;
 
   if (transitions_ini && !transitions_ini_is_dirty)
     return transitions_ini;
-  g_debug("%s transitions.ini",
-          transitions_ini_is_dirty ? "reloading" : "loading");
+
+  /* Check for a file in the theme directory first, otherwise
+   * use the one in the hildon-desktop dir */
+  fname = g_file_test(TRANSITIONS_INI_FROM_THEME, G_FILE_TEST_EXISTS)
+    ? TRANSITIONS_INI_FROM_THEME : TRANSITIONS_INI;
+  g_debug("%s: %s %s", __FUNCTION__,
+          transitions_ini_is_dirty ? "reloading" : "loading", fname);
 
   error = NULL;
   ini = g_key_file_new();
-  if (!g_key_file_load_from_file (ini, TRANSITIONS_INI, 0, &error))
+  if (!g_key_file_load_from_file (ini, fname, 0, &error))
     { /* Use the previous @transitions_ini. */
-      g_warning("couldn't load %s: %s", TRANSITIONS_INI, error->message);
+      g_warning("%s: couldn't load %s: %s", __FUNCTION__,
+                fname, error->message);
       if (!transitions_ini)
-        g_warning("using default settings");
+        g_warning("%s: using default settings", __FUNCTION__);
       g_error_free(error);
       g_key_file_free(ini);
       return transitions_ini;
     }
 
   /* Use the new @transitions_ini. */
+  if (transitions_ini)
+    g_key_file_free(transitions_ini);
   transitions_ini = ini;
 
-  if (!transitions_ini_watcher)
+  if (!transitions_ini_watcher || transitions_ini_is_dirty > TRUE)
     {
       static int inofd = -1, watch = -1;
 
       /* Create an inotify if we haven't. */
-      if (inofd < 0 && (inofd = inotify_init()) < 0)
+      if (inofd < 0)
         {
-          g_warning("inotify_init: %s", strerror(errno));
-          goto out;
-        }
+          if ((inofd = inotify_init()) < 0)
+            {
+              g_warning("inotify_init: %s", strerror(errno));
+              goto out;
+            }
 
-      if (watch >= 0)
+          g_assert(!transitions_ini_watcher);
+          transitions_ini_watcher = g_io_channel_unix_new(inofd);
+          g_io_channel_set_encoding(transitions_ini_watcher, NULL, NULL);
+          g_io_add_watch(transitions_ini_watcher, G_IO_IN,
+                         transitions_ini_changed, NULL);
+        }
+      else if (watch >= 0)
         /* Remove the previous watch. */
         inotify_rm_watch(inofd, watch);
-      watch = inotify_add_watch(inofd, TRANSITIONS_INI,
-                              IN_MODIFY | IN_MOVE_SELF | IN_DELETE_SELF);
+
+      g_assert(transitions_ini_watcher != NULL);
+      watch = inotify_add_watch(inofd, fname,
+                                IN_MODIFY|IN_MOVE_SELF|IN_DELETE_SELF);
       if (watch < 0)
         {
           g_warning("inotify_add_watch: %s", strerror(errno));
           goto out;
         }
 
-      transitions_ini_watcher = g_io_channel_unix_new(inofd);
-      g_io_add_watch_full(transitions_ini_watcher, G_PRIORITY_DEFAULT, G_IO_IN,
-                          transitions_ini_changed, &transitions_ini_watcher,
-                          (GDestroyNotify)g_nullify_pointer);
-      g_debug("watching transitions.ini");
+      g_debug("watching %s", fname);
     }
 
   /* Stop reloading @transitions_ini if we can watch it. */
@@ -1803,7 +2024,7 @@ hd_transition_get_int(const gchar *transition, const char *key,
   value = g_key_file_get_integer(ini, transition, key, &error);
   if (error)
     {
-      g_warning("couldn't read %s::%s from transitions.ini: %s",
+      g_debug("couldn't read int %s::%s from transitions.ini: %s",
                 transition, key, error->message);
       g_error_free(error);
       return default_val;
@@ -1827,7 +2048,7 @@ hd_transition_get_double(const gchar *transition,
   value = g_key_file_get_double (ini, transition, key, &error);
   if (error)
     {
-      g_warning("couldn't read %s::%s from transitions.ini: %s",
+      g_debug("couldn't read double %s::%s from transitions.ini: %s",
                 transition, key, error->message);
       g_error_free(error);
       return default_val;
@@ -1835,3 +2056,38 @@ hd_transition_get_double(const gchar *transition,
 
   return value;
 }
+
+/* Returns a newly-allocated string that must *always* be freed by the caller */
+gchar *
+hd_transition_get_string(const gchar *transition, const char *key,
+                      gchar *default_val)
+{
+  gchar *value;
+  GError *error;
+  GKeyFile *ini;
+
+  if (!(ini = hd_transition_get_keyfile()))
+    return default_val;
+
+  error = NULL;
+  value = g_key_file_get_string (ini, transition, key, &error);
+  if (error)
+    {
+      g_debug("couldn't read string %s::%s from transitions.ini: %s",
+                transition, key, error->message);
+      g_error_free(error);
+      return g_strdup(default_val);
+    }
+
+  return value;
+}
+
+HdKeyFrameList *
+hd_transition_get_keyframes(const gchar *transition, const char *key,
+                            gchar *default_val)
+{
+  char *keyframetext = hd_transition_get_string(transition, key, default_val);
+  HdKeyFrameList *keyframes = hd_key_frame_list_create(keyframetext);
+  g_free(keyframetext);
+  return keyframes;
+}
